"""Smart agent implementation with built-in tools."""

from __future__ import annotations

import asyncio
import base64
import shlex
import sys
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, Protocol, cast

from mcp.types import BlobResourceContents, ReadResourceResult, TextResourceContents

from fast_agent.agents.agent_types import AgentConfig, AgentType
from fast_agent.agents.mcp_agent import McpAgent
from fast_agent.agents.workflow.agents_as_tools_agent import (
    AgentsAsToolsAgent,
    AgentsAsToolsOptions,
)
from fast_agent.commands.command_catalog import (
    command_action_names,
    command_alias_map,
    get_command_spec,
    suggest_command_name,
)
from fast_agent.commands.command_discovery import (
    command_discovery_names,
    parse_commands_discovery_arguments,
    render_command_detail_markdown,
    render_commands_index_markdown,
    render_commands_json,
)
from fast_agent.commands.context import (
    CommandContext,
    NonInteractiveCommandIOBase,
    StaticAgentProvider,
)
from fast_agent.commands.handlers import cards_manager as cards_handlers
from fast_agent.commands.handlers import display as display_handlers
from fast_agent.commands.handlers import mcp_runtime as mcp_runtime_handlers
from fast_agent.commands.handlers import model as model_handlers
from fast_agent.commands.handlers import models_manager as models_handlers
from fast_agent.commands.handlers import prompts as prompt_handlers
from fast_agent.commands.handlers import sessions as sessions_handlers
from fast_agent.commands.handlers import skills as skills_handlers
from fast_agent.commands.handlers import tools as tools_handlers
from fast_agent.commands.handlers.shared import clear_agent_histories
from fast_agent.commands.renderers.command_markdown import render_command_outcome_markdown
from fast_agent.commands.results import CommandMessage, CommandOutcome
from fast_agent.core.agent_app import AgentApp
from fast_agent.core.agent_card_loader import load_agent_cards
from fast_agent.core.agent_card_validation import AgentCardScanResult, scan_agent_card_path
from fast_agent.core.default_agent import resolve_default_agent_name
from fast_agent.core.direct_factory import (
    create_basic_agents_in_dependency_order,
    get_model_factory,
)
from fast_agent.core.exceptions import AgentConfigError
from fast_agent.core.instruction_utils import apply_instruction_context
from fast_agent.core.internal_resources import (
    InternalResource,
    get_internal_resource,
    list_internal_resources,
    read_internal_resource,
)
from fast_agent.core.logging.logger import get_logger
from fast_agent.core.prompt_templates import enrich_with_environment_context
from fast_agent.core.validation import validate_provider_keys_post_creation
from fast_agent.mcp.helpers.content_helpers import get_text
from fast_agent.mcp.prompts.prompt_load import load_prompt
from fast_agent.mcp.ui_mixin import McpUIMixin
from fast_agent.paths import resolve_environment_paths
from fast_agent.tools.function_tool_loader import build_default_function_tool

if TYPE_CHECKING:
    from fast_agent.agents.llm_agent import LlmAgent
    from fast_agent.config import MCPServerSettings
    from fast_agent.context import Context
    from fast_agent.core.agent_card_types import AgentCardData
    from fast_agent.interfaces import AgentProtocol
    from fast_agent.mcp.mcp_aggregator import MCPAttachOptions, MCPAttachResult, MCPDetachResult

logger = get_logger(__name__)

_INTERNAL_RESOURCE_SERVER = "internal"


@dataclass(frozen=True)
class _SmartCardBundle:
    agents_dict: dict[str, "AgentCardData | dict[str, Any]"]
    message_files: dict[str, list[Path]]


class _McpCapableAgent(Protocol):
    async def attach_mcp_server(
        self,
        *,
        server_name: str,
        server_config: MCPServerSettings | None = None,
        options: MCPAttachOptions | None = None,
    ) -> MCPAttachResult: ...

    async def detach_mcp_server(self, server_name: str) -> MCPDetachResult: ...

    def list_attached_mcp_servers(self) -> list[str]: ...


@dataclass(frozen=True)
class _SmartConnectSummary:
    connected: list[str]
    warnings: list[str]


class _SmartToolMcpManager:
    """Minimal MCP runtime manager adapter for temporary smart-tool agents."""

    def __init__(
        self,
        agents: Mapping[str, AgentProtocol],
        configured_server_names: set[str],
    ) -> None:
        self._agents = agents
        self._configured_server_names = configured_server_names

    def _agent(self, name: str) -> _McpCapableAgent:
        agent = self._agents.get(name)
        if agent is None:
            raise AgentConfigError("Unknown agent", f"Agent '{name}' is not loaded")

        required = ("attach_mcp_server", "detach_mcp_server", "list_attached_mcp_servers")
        if not all(callable(getattr(agent, attr, None)) for attr in required):
            raise AgentConfigError(
                "Agent does not support runtime MCP connection",
                f"Agent '{name}' cannot attach MCP servers",
            )

        return agent  # type: ignore[return-value]

    async def attach_mcp_server(
        self,
        agent_name: str,
        server_name: str,
        server_config: MCPServerSettings | None = None,
        options: MCPAttachOptions | None = None,
    ) -> MCPAttachResult:
        agent = self._agent(agent_name)
        return await agent.attach_mcp_server(
            server_name=server_name,
            server_config=server_config,
            options=options,
        )

    async def detach_mcp_server(self, agent_name: str, server_name: str) -> MCPDetachResult:
        agent = self._agent(agent_name)
        return await agent.detach_mcp_server(server_name)

    async def list_attached_mcp_servers(self, agent_name: str) -> list[str]:
        agent = self._agent(agent_name)
        return list(agent.list_attached_mcp_servers())

    async def list_configured_detached_mcp_servers(self, agent_name: str) -> list[str]:
        attached = set(await self.list_attached_mcp_servers(agent_name))
        return sorted(self._configured_server_names - attached)


@dataclass(slots=True)
class _SmartToolCommandIO(NonInteractiveCommandIOBase):
    """Non-interactive command IO that buffers emitted messages."""

    messages: list[CommandMessage]

    async def emit(self, message: CommandMessage) -> None:
        self.messages.append(message)


def _resolve_default_agent_name(
    agents: Mapping[str, AgentProtocol],
    *,
    tool_only_agents: set[str],
) -> str:
    default_agent_name = resolve_default_agent_name(
        agents,
        is_default=lambda _name, agent: bool(getattr(agent.config, "default", False)),
        is_tool_only=lambda name, _agent: name in tool_only_agents,
    )
    if default_agent_name is None:
        raise AgentConfigError("Smart tool requires at least one agent")
    return default_agent_name


def _collect_outcome_messages(outcome: "CommandOutcome") -> tuple[list[str], list[str]]:
    errors: list[str] = []
    warnings: list[str] = []
    for message in outcome.messages:
        text = str(message.text)
        if message.channel == "error":
            errors.append(text)
        elif message.channel == "warning":
            warnings.append(text)
    return errors, warnings


def _resolve_command_agent_map(agent: Any) -> dict[str, object]:
    registry = getattr(agent, "agent_registry", None)
    agents: dict[str, object] = {}
    if isinstance(registry, Mapping):
        agents.update({str(name): member for name, member in registry.items()})

    agent_name = str(getattr(agent, "name", ""))
    if agent_name:
        agents.setdefault(agent_name, agent)
    return agents


def _build_command_context(agent: Any) -> tuple[CommandContext, _SmartToolCommandIO]:
    context = getattr(agent, "context", None)
    settings = context.config if context else None
    agent_name = str(getattr(agent, "name", ""))
    if not agent_name:
        raise AgentConfigError("Command execution requires named agent", "Agent has no name")

    io = _SmartToolCommandIO(messages=[])
    provider = StaticAgentProvider(_resolve_command_agent_map(agent))
    return (
        CommandContext(
            agent_provider=provider,
            current_agent_name=agent_name,
            io=io,
            settings=settings,
        ),
        io,
    )


def _render_command_outcome(
    outcome: CommandOutcome,
    *,
    heading: str,
    io: _SmartToolCommandIO,
) -> str:
    return render_command_outcome_markdown(
        outcome,
        heading=heading,
        extra_messages=io.messages,
    )


async def _run_check_subprocess(argument: str | None) -> CommandOutcome:
    command = [
        sys.executable,
        "-m",
        "fast_agent.cli.main",
        "--no-color",
        "check",
    ]
    if argument and argument.strip():
        try:
            command.extend(shlex.split(argument))
        except ValueError as exc:
            raise AgentConfigError("Invalid check arguments", str(exc)) from exc

    process = await asyncio.create_subprocess_exec(
        *command,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    stdout, stderr = await process.communicate()

    outcome = CommandOutcome()
    stdout_text = stdout.decode("utf-8", errors="replace").strip()
    stderr_text = stderr.decode("utf-8", errors="replace").strip()
    if stdout_text:
        outcome.add_message(stdout_text)
    if stderr_text:
        channel = "error" if process.returncode else "warning"
        outcome.add_message(stderr_text, channel=channel)
    if process.returncode != 0 and not stderr_text:
        outcome.add_message(
            f"fast-agent check exited with status code {process.returncode}.",
            channel="error",
        )
    return outcome


async def _run_named_command_call(
    agent: Any,
    *,
    command_name: str,
    action: str | None = None,
    argument: str | None = None,
    heading: str | None = None,
) -> str:
    normalized_command = command_name.strip().lower()
    normalized_action = (action or "").strip().lower()

    if normalized_command == "check":
        outcome = await _run_check_subprocess(argument)
        return render_command_outcome_markdown(outcome, heading=heading or "check")

    context, io = _build_command_context(agent)
    agent_name = context.current_agent_name

    if normalized_command == "skills":
        selected_action = normalized_action or "list"
        outcome = await skills_handlers.handle_skills_command(
            context,
            agent_name=agent_name,
            action=selected_action,
            argument=argument,
        )
        final_heading = heading or f"skills.{selected_action}"
        return _render_command_outcome(outcome, heading=final_heading, io=io)

    if normalized_command == "cards":
        selected_action = normalized_action or "list"
        outcome = await cards_handlers.handle_cards_command(
            context,
            agent_name=agent_name,
            action=selected_action,
            argument=argument,
        )
        final_heading = heading or f"cards.{selected_action}"
        return _render_command_outcome(outcome, heading=final_heading, io=io)

    if normalized_command == "model":
        selected_action = normalized_action or "reasoning"
        management_actions = {"doctor", "references", "catalog", "help"}
        if selected_action in management_actions:
            outcome = await models_handlers.handle_models_command(
                context,
                agent_name=agent_name,
                action=selected_action,
                argument=argument,
            )
            final_heading = heading or f"model.{selected_action}"
            return _render_command_outcome(outcome, heading=final_heading, io=io)

        if selected_action == "verbosity":
            outcome = await model_handlers.handle_model_verbosity(
                context,
                agent_name=agent_name,
                value=argument,
            )
        elif selected_action == "fast":
            outcome = await model_handlers.handle_model_fast(
                context,
                agent_name=agent_name,
                value=argument,
            )
        elif selected_action == "web_search":
            outcome = await model_handlers.handle_model_web_search(
                context,
                agent_name=agent_name,
                value=argument,
            )
        elif selected_action == "web_fetch":
            outcome = await model_handlers.handle_model_web_fetch(
                context,
                agent_name=agent_name,
                value=argument,
            )
        elif selected_action == "switch":
            outcome = await model_handlers.handle_model_switch(
                context,
                agent_name=agent_name,
                value=argument,
            )
            if outcome.reset_session:
                if not context.noenv:
                    outcome.add_message(
                        "Model switch starts a new session to avoid mixing histories.",
                        channel="info",
                    )
                    session_outcome = await sessions_handlers.handle_create_session(
                        context,
                        session_name=None,
                    )
                    outcome.messages.extend(session_outcome.messages)
                else:
                    outcome.add_message(
                        "Model switch cleared in-memory history (--noenv disables session persistence).",
                        channel="info",
                    )
                cleared = clear_agent_histories(_resolve_command_agent_map(agent))
                if cleared:
                    outcome.add_message(
                        f"Cleared agent history: {', '.join(sorted(cleared))}",
                        channel="info",
                    )
        else:
            outcome = await model_handlers.handle_model_reasoning(
                context,
                agent_name=agent_name,
                value=argument,
            )
        final_heading = heading or f"model.{selected_action}"
        return _render_command_outcome(outcome, heading=final_heading, io=io)

    raise AgentConfigError(
        "Unsupported command",
        f"Command '{command_name}' is not supported by slash_command.",
    )


def _parse_family_command_action(
    command_name: str,
    arguments: str,
) -> tuple[str, str | None]:
    spec = get_command_spec(command_name)
    if spec is None:
        raise AgentConfigError(
            "Unsupported command",
            f"Command '{command_name}' is not supported by slash_command.",
        )

    trimmed = arguments.strip()
    if not trimmed:
        return spec.default_action, None

    try:
        tokens = shlex.split(trimmed)
    except ValueError as exc:
        raise AgentConfigError("Invalid command arguments", str(exc)) from exc

    if not tokens:
        return spec.default_action, None

    supported_actions = set(command_action_names(command_name))
    first = tokens[0].lower()

    if first in {"help", "--help", "-h"}:
        return "help", None

    aliases = command_alias_map(command_name)
    alias_resolved = aliases.get(first)
    if alias_resolved is not None:
        remainder = " ".join(tokens[1:]).strip() or None
        return alias_resolved, remainder

    if command_name == "skills":
        extra_actions = {
            "available",
            "search",
        }
        if first in extra_actions:
            remainder = " ".join(tokens[1:]).strip() or None
            return first, remainder

    if first in supported_actions:
        remainder = " ".join(tokens[1:]).strip() or None
        return first, remainder

    return spec.default_action, trimmed


def _parse_slash_command_text(command: str) -> tuple[str, str]:
    text = command.strip()
    if not text:
        raise AgentConfigError(
            "Slash command is empty",
            "Pass a command like '/mcp list' or '/skills list'.",
        )

    if text.startswith("/"):
        text = text[1:].lstrip()

    if not text:
        raise AgentConfigError(
            "Slash command is empty",
            "Pass a command like '/mcp list' or '/skills list'.",
        )

    command_name, _, arguments = text.partition(" ")
    return command_name.strip().lower(), arguments.strip()


def _smart_slash_usage() -> str:
    return render_commands_index_markdown(command_names=command_discovery_names())


def _render_unknown_slash_command(command_name: str) -> str:
    suggestions = suggest_command_name(command_name)
    suggestion_line = ""
    if suggestions:
        suggestion_line = "\nDid you mean: " + ", ".join(f"`/{name}`" for name in suggestions)

    return (
        f"Unknown slash command '/{command_name}'."
        f"{suggestion_line}\n\n"
        f"{_smart_slash_usage()}"
    )


def _mcp_usage_text() -> str:
    return (
        "Usage: /mcp [list|connect|disconnect|reconnect|session|help] [args]\n"
        "- /mcp list\n"
        "- /mcp connect <target> [--name <server>] [--auth <token-value>] [--timeout <seconds>]\n"
        "- /mcp session [list|jar|new|use|clear] [args]"
    )


def _model_usage_text() -> str:
    return (
        "Usage: /model "
        "[reasoning|verbosity|fast|web_search|web_fetch|switch|doctor|references|catalog|help] [args]"
    )


def _run_commands_slash_command_call(arguments: str) -> str:
    try:
        request = parse_commands_discovery_arguments(arguments)
    except ValueError as exc:
        raise AgentConfigError("Invalid /commands arguments", str(exc)) from exc

    if request.as_json:
        return render_commands_json(command_name=request.command_name)

    if request.command_name is None:
        return render_commands_index_markdown(command_names=command_discovery_names())

    detail = render_command_detail_markdown(request.command_name)
    if detail is not None:
        return detail

    suggestions = suggest_command_name(request.command_name)
    suggestion_text = ""
    if suggestions:
        suggestion_text = "\nDid you mean: " + ", ".join(f"`/{name}`" for name in suggestions)
    return (
        f"Unknown command family: `{request.command_name}`.\n"
        f"Use `/commands` to list available commands.{suggestion_text}"
    )


def _render_smart_slash_outcome(
    outcome: CommandOutcome,
    *,
    heading: str,
    io: _SmartToolCommandIO,
) -> str:
    return _render_command_outcome(outcome, heading=heading, io=io)


def _parse_mcp_session_args(tokens: list[str]) -> tuple[str, str | None, str | None, str | None, bool]:
    session_tokens = tokens[1:]
    action = "list"
    server_identity: str | None = None
    session_id: str | None = None
    title: str | None = None
    clear_all = False

    if not session_tokens:
        return action, server_identity, session_id, title, clear_all

    action = session_tokens[0].lower()
    args = session_tokens[1:]

    if action == "list":
        if len(args) > 1:
            raise AgentConfigError(
                "Invalid /mcp session arguments",
                "Usage: /mcp session list [<server_or_identity>]",
            )
        server_identity = args[0] if args else None
        return action, server_identity, session_id, title, clear_all

    if action == "jar":
        if len(args) > 1:
            raise AgentConfigError(
                "Invalid /mcp session arguments",
                "Usage: /mcp session jar [<server_or_identity>]",
            )
        server_identity = args[0] if args else None
        return action, server_identity, session_id, title, clear_all

    if action in {"new", "create"}:
        idx = 0
        while idx < len(args):
            token = args[idx]
            if token == "--title":
                idx += 1
                if idx >= len(args):
                    raise AgentConfigError("Invalid /mcp session arguments", "Missing value for --title")
                title = args[idx]
            elif token.startswith("--title="):
                title = token.split("=", 1)[1] or None
                if title is None:
                    raise AgentConfigError(
                        "Invalid /mcp session arguments",
                        "Missing value for --title",
                    )
            elif token.startswith("--"):
                raise AgentConfigError(
                    "Invalid /mcp session arguments",
                    f"Unknown flag: {token}",
                )
            elif server_identity is None:
                server_identity = token
            else:
                raise AgentConfigError(
                    "Invalid /mcp session arguments",
                    f"Unexpected argument: {token}",
                )
            idx += 1
        return "new", server_identity, session_id, title, clear_all

    if action in {"resume", "use"}:
        if len(args) != 2:
            raise AgentConfigError(
                "Invalid /mcp session arguments",
                "Usage: /mcp session use <server_or_identity> <session_id>",
            )
        server_identity, session_id = args
        return "use", server_identity, session_id, title, clear_all

    if action == "clear":
        for token in args:
            if token == "--all":
                clear_all = True
                continue
            if token.startswith("--"):
                raise AgentConfigError(
                    "Invalid /mcp session arguments",
                    f"Unknown flag: {token}",
                )
            if server_identity is None:
                server_identity = token
            else:
                raise AgentConfigError(
                    "Invalid /mcp session arguments",
                    f"Unexpected argument: {token}",
                )

        if clear_all and server_identity is not None:
            raise AgentConfigError(
                "Invalid /mcp session arguments",
                "Use either --all or a specific server, not both",
            )

        if not clear_all and server_identity is None:
            clear_all = True

        return action, server_identity, session_id, title, clear_all

    if args:
        raise AgentConfigError(
            "Invalid /mcp session arguments",
            (
                "Usage: /mcp session [list [server]|jar [server]|new [server] "
                "[--title <title>]|use <server> <session_id>|clear [server|--all]]"
            ),
        )

    server_identity = action
    return "list", server_identity, session_id, title, clear_all


async def _run_mcp_slash_command_call(agent: Any, arguments: str) -> str:
    context, io = _build_command_context(agent)
    agent_name = context.current_agent_name
    runtime_manager = _SmartToolMcpManager(
        {agent_name: agent},
        configured_server_names=_context_server_names(getattr(agent, "context", None)),
    )

    args = arguments.strip() or "list"
    try:
        tokens = shlex.split(args)
    except ValueError as exc:
        raise AgentConfigError("Invalid /mcp arguments", str(exc)) from exc

    if not tokens:
        tokens = ["list"]

    subcmd = tokens[0].lower()

    if subcmd in {"help", "--help", "-h"}:
        return _mcp_usage_text()

    if subcmd == "list":
        outcome = await mcp_runtime_handlers.handle_mcp_list(
            context,
            manager=runtime_manager,
            agent_name=agent_name,
        )
        return _render_smart_slash_outcome(outcome, heading="mcp", io=io)

    if subcmd == "connect":
        if len(tokens) < 2:
            raise AgentConfigError(
                "Invalid /mcp connect arguments",
                (
                    "Usage: /mcp connect <target> [--name <server>] [--auth <token-value>] "
                    "[--timeout <seconds>] [--oauth|--no-oauth] [--reconnect|--no-reconnect]"
                ),
            )

        target_text = " ".join(tokens[1:])
        outcome = await mcp_runtime_handlers.handle_mcp_connect(
            context,
            manager=runtime_manager,
            agent_name=agent_name,
            target_text=target_text,
        )
        return _render_smart_slash_outcome(outcome, heading="mcp", io=io)

    if subcmd == "disconnect":
        if len(tokens) != 2:
            raise AgentConfigError(
                "Invalid /mcp disconnect arguments",
                "Usage: /mcp disconnect <server_name>",
            )

        outcome = await mcp_runtime_handlers.handle_mcp_disconnect(
            context,
            manager=runtime_manager,
            agent_name=agent_name,
            server_name=tokens[1],
        )
        return _render_smart_slash_outcome(outcome, heading="mcp", io=io)

    if subcmd == "reconnect":
        if len(tokens) != 2:
            raise AgentConfigError(
                "Invalid /mcp reconnect arguments",
                "Usage: /mcp reconnect <server_name>",
            )

        outcome = await mcp_runtime_handlers.handle_mcp_reconnect(
            context,
            manager=runtime_manager,
            agent_name=agent_name,
            server_name=tokens[1],
        )
        return _render_smart_slash_outcome(outcome, heading="mcp", io=io)

    if subcmd == "session":
        action, server_identity, session_id, title, clear_all = _parse_mcp_session_args(tokens)
        outcome = await mcp_runtime_handlers.handle_mcp_session(
            context,
            agent_name=agent_name,
            action=cast("mcp_runtime_handlers.McpSessionAction", action),
            server_identity=server_identity,
            session_id=session_id,
            title=title,
            clear_all=clear_all,
        )
        return _render_smart_slash_outcome(outcome, heading="mcp", io=io)

    raise AgentConfigError(
        "Unsupported /mcp subcommand",
        _mcp_usage_text(),
    )


async def _run_slash_command_call(agent: Any, command: str) -> str:
    command_name, arguments = _parse_slash_command_text(command)

    if command_name in {"help", "?"}:
        return _smart_slash_usage()

    if command_name == "commands":
        return _run_commands_slash_command_call(arguments)

    if command_name in {"skills", "cards", "model"}:
        action, argument = _parse_family_command_action(command_name, arguments)
        normalized_action = action

        heading = f"{command_name}.{normalized_action}"
        return await _run_named_command_call(
            agent,
            command_name=command_name,
            action=normalized_action,
            argument=argument,
            heading=heading,
        )

    if command_name == "check":
        argument = arguments.strip() or None
        return await _run_named_command_call(
            agent,
            command_name="check",
            action="run",
            argument=argument,
            heading="check",
        )

    if command_name == "mcp":
        return await _run_mcp_slash_command_call(agent, arguments)

    context, io = _build_command_context(agent)
    agent_name = context.current_agent_name

    if command_name == "tools":
        outcome = await tools_handlers.handle_list_tools(context, agent_name=agent_name)
        return _render_smart_slash_outcome(outcome, heading="tools", io=io)

    if command_name == "prompts":
        outcome = await prompt_handlers.handle_list_prompts(context, agent_name=agent_name)
        return _render_smart_slash_outcome(outcome, heading="prompts", io=io)

    if command_name == "usage":
        outcome = await display_handlers.handle_show_usage(context, agent_name=agent_name)
        return _render_smart_slash_outcome(outcome, heading="usage", io=io)

    if command_name == "system":
        outcome = await display_handlers.handle_show_system(context, agent_name=agent_name)
        return _render_smart_slash_outcome(outcome, heading="system", io=io)

    if command_name == "markdown":
        outcome = await display_handlers.handle_show_markdown(context, agent_name=agent_name)
        return _render_smart_slash_outcome(outcome, heading="markdown", io=io)

    if command_name in {"mcpstatus", "status"}:
        outcome = await display_handlers.handle_show_mcp_status(context, agent_name=agent_name)
        return _render_smart_slash_outcome(outcome, heading="mcpstatus", io=io)

    return _render_unknown_slash_command(command_name)


def _context_server_names(context: "Context | None") -> set[str]:
    if context and context.config and context.config.mcp and context.config.mcp.servers:
        return set(context.config.mcp.servers.keys())
    return set()


async def _apply_runtime_mcp_connections(
    *,
    context: "Context | None",
    agents_map: Mapping[str, AgentProtocol],
    target_agent_name: str,
    mcp_connect: Sequence[str],
) -> _SmartConnectSummary:
    configured_names = _context_server_names(context)

    manager = _SmartToolMcpManager(agents_map, configured_server_names=configured_names)
    connected_names: list[str] = []
    warnings: list[str] = []

    for raw_target in mcp_connect:
        target = raw_target.strip()
        if not target:
            continue

        outcome = await mcp_runtime_handlers.handle_mcp_connect(
            None,
            manager=manager,
            agent_name=target_agent_name,
            target_text=target,
        )
        errors, target_warnings = _collect_outcome_messages(outcome)
        warnings.extend(target_warnings)
        if errors:
            raise AgentConfigError(
                "Failed to connect MCP server for smart tool call",
                "\n".join(errors),
            )

        parsed = mcp_runtime_handlers.parse_connect_input(target)
        mode = mcp_runtime_handlers.infer_connect_mode(parsed.target_text)
        resolved_name = parsed.server_name or mcp_runtime_handlers.infer_server_name(
            parsed.target_text,
            mode,
        )
        connected_names.append(resolved_name)

    return _SmartConnectSummary(connected=connected_names, warnings=warnings)


def _resolve_agent_card_path(path_value: str, context: Context | None) -> Path:
    candidate = Path(path_value).expanduser()
    if candidate.is_absolute():
        if candidate.exists():
            return candidate.resolve()
    else:
        cwd_candidate = (Path.cwd() / candidate).resolve()
        if cwd_candidate.exists():
            return cwd_candidate

    env_paths = resolve_environment_paths(
        settings=context.config if context else None, cwd=Path.cwd()
    )
    for base in (env_paths.agent_cards, env_paths.tool_cards):
        env_candidate = (base / candidate).resolve()
        if env_candidate.exists():
            return env_candidate

    raise AgentConfigError(
        "AgentCard path not found",
        f"Tried: {candidate} (cwd), {env_paths.agent_cards}, {env_paths.tool_cards}",
    )


def _ensure_basic_only_cards(cards: Sequence) -> _SmartCardBundle:
    agents_dict: dict[str, "AgentCardData | dict[str, Any]"] = {}
    message_files: dict[str, list[Path]] = {}
    for card in cards:
        agent_data = card.agent_data
        agent_type = agent_data.get("type")
        if agent_type != AgentType.BASIC.value:
            raise AgentConfigError(
                "Smart tool only supports 'agent' cards",
                f"Card '{card.name}' has unsupported type '{agent_type}'",
            )
        agents_dict[card.name] = agent_data
        if card.message_files:
            message_files[card.name] = list(card.message_files)
    return _SmartCardBundle(agents_dict=agents_dict, message_files=message_files)


def _apply_agent_card_histories(
    agents: dict[str, AgentProtocol],
    message_map: Mapping[str, list[Path]],
) -> None:
    for name, history_files in message_map.items():
        agent = agents.get(name)
        if agent is None:
            continue
        messages = []
        for history_file in history_files:
            messages.extend(load_prompt(history_file))
        agent.clear(clear_prompts=True)
        agent.message_history.extend(messages)


async def _apply_instruction_context(
    agents: dict[str, AgentProtocol],
    context_vars: Mapping[str, str],
) -> None:
    await apply_instruction_context(agents.values(), context_vars)


async def _shutdown_agents(agents: Mapping[str, AgentProtocol]) -> None:
    for agent in agents.values():
        try:
            await agent.shutdown()
        except Exception:
            pass


def _format_validation_results(results: Sequence[AgentCardScanResult]) -> str:
    if not results:
        return "No AgentCards found."

    lines: list[str] = []
    for entry in results:
        if entry.ignored_reason:
            status = f"ignored - {entry.ignored_reason}"
        elif entry.errors:
            status = "error"
        else:
            status = "ok"
        lines.append(f"{entry.name} ({entry.type}) - {status}")
        for error in entry.errors:
            lines.append(f"  - {error}")
    return "\n".join(lines)


async def _create_smart_tool_agents(
    context: "Context | None",
    agent_card_path: str,
) -> tuple[dict[str, AgentProtocol], _SmartCardBundle, str, set[str]]:
    if context is None:
        raise AgentConfigError("Smart tool requires an initialized context")

    resolved_path = _resolve_agent_card_path(agent_card_path, context)
    cards = load_agent_cards(resolved_path)
    bundle = _ensure_basic_only_cards(cards)

    def model_factory_func(model=None, request_params=None):
        return get_model_factory(context, model=model, request_params=request_params)

    agents_map = await create_basic_agents_in_dependency_order(
        context,
        bundle.agents_dict,
        model_factory_func,
    )
    validate_provider_keys_post_creation(agents_map)

    tool_only_agents = {
        name for name, data in bundle.agents_dict.items() if data.get("tool_only", False)
    }
    default_agent_name = _resolve_default_agent_name(
        agents_map,
        tool_only_agents=tool_only_agents,
    )

    return agents_map, bundle, default_agent_name, tool_only_agents


async def _hydrate_smart_agents_for_execution(
    agents_map: dict[str, AgentProtocol],
    bundle: _SmartCardBundle,
) -> None:
    if bundle.message_files:
        _apply_agent_card_histories(agents_map, bundle.message_files)

    context_vars: dict[str, str] = {}
    enrich_with_environment_context(
        context_vars,
        str(Path.cwd()),
        {"name": "fast-agent"},
    )
    await _apply_instruction_context(agents_map, context_vars)


def _extract_read_resource_text(result: ReadResourceResult, *, max_chars: int = 4000) -> str:
    lines: list[str] = []
    for idx, content in enumerate(result.contents, start=1):
        if isinstance(content, TextResourceContents):
            text = content.text
            lines.append(f"[{idx}] text ({content.mimeType or 'unknown'})")
            lines.append(text)
            continue

        if isinstance(content, BlobResourceContents):
            blob_len = len(content.blob)
            preview = ""
            try:
                decoded = base64.b64decode(content.blob)
                preview = decoded[:400].decode("utf-8", errors="replace")
            except Exception:
                preview = "<binary blob>"
            lines.append(f"[{idx}] blob ({content.mimeType or 'unknown'}, {blob_len} b64 chars)")
            if preview:
                lines.append(preview)
            continue

        text = get_text(content)
        if text:
            lines.append(f"[{idx}] content")
            lines.append(text)

    joined = "\n".join(lines).strip()
    if len(joined) <= max_chars:
        return joined
    return joined[: max_chars - 1] + "…\n[truncated]"


def _format_smart_resource_listing(
    resources: Mapping[str, list[str]],
    templates: Mapping[str, Sequence[Any]],
    *,
    mcp_server_names: Sequence[str] = (),
    internal_resources: Sequence[InternalResource] = (),
) -> str:
    lines: list[str] = []
    server_names = sorted(
        (set(resources.keys()) | set(templates.keys()) | set(mcp_server_names))
        - {_INTERNAL_RESOURCE_SERVER}
    )
    if not server_names and not internal_resources:
        return "No resources available."

    lines.append("server_names:")
    lines.append(f"  - {_INTERNAL_RESOURCE_SERVER} (always available)")
    if server_names:
        for server_name in server_names:
            lines.append(f"  - {server_name}")
    else:
        lines.append("  - (no attached MCP servers)")
    lines.append("")

    if internal_resources:
        lines.append(f"[{_INTERNAL_RESOURCE_SERVER}]")
        lines.append("resources:")
        for resource in internal_resources:
            lines.append(f"  - {resource.uri}")
            lines.append(f"    description: {resource.description}")
            lines.append(f"    why: {resource.why}")
        lines.append("templates: []")
        if server_names:
            lines.append("")

    for server_name in server_names:
        lines.append(f"[{server_name}]")
        server_resources = resources.get(server_name, [])
        if server_resources:
            lines.append("resources:")
            for uri in server_resources:
                lines.append(f"  - {uri}")
        else:
            lines.append("resources: []")

        server_templates = templates.get(server_name, [])
        if server_templates:
            lines.append("templates:")
            for template in server_templates:
                uri_template = getattr(template, "uriTemplate", "")
                name = getattr(template, "name", "")
                if name:
                    lines.append(f"  - {name}: {uri_template}")
                else:
                    lines.append(f"  - {uri_template}")
        else:
            lines.append("templates: []")

    return "\n".join(lines)


def _include_internal_resources(server_name: str | None) -> bool:
    return server_name is None or server_name == _INTERNAL_RESOURCE_SERVER


def _include_mcp_resources(server_name: str | None) -> bool:
    return server_name != _INTERNAL_RESOURCE_SERVER


def _is_internal_resource_uri(uri: str) -> bool:
    return uri.strip().startswith("internal://")


async def _run_internal_resource_read_call(uri: str) -> str:
    resource = get_internal_resource(uri)
    content = read_internal_resource(resource.uri)
    header = f"Resource: {resource.uri}\nTitle: {resource.title}"
    return f"{header}\n\n{content}" if content else header


async def _run_smart_call(
    context: "Context | None",
    agent_card_path: str,
    message: str,
    *,
    mcp_connect: Sequence[str] | None = None,
    disable_streaming: bool = False,
) -> str:
    agents_map, bundle, default_agent_name, tool_only_agents = await _create_smart_tool_agents(
        context,
        agent_card_path,
    )
    try:
        app = AgentApp(agents_map, tool_only_agents=tool_only_agents)

        if mcp_connect:
            connect_summary = await _apply_runtime_mcp_connections(
                context=context,
                agents_map=agents_map,
                target_agent_name=default_agent_name,
                mcp_connect=mcp_connect,
            )
            if connect_summary.connected:
                logger.info(
                    "Connected runtime MCP servers for smart tool call",
                    data={
                        "agent": default_agent_name,
                        "servers": connect_summary.connected,
                    },
                )
            for warning in connect_summary.warnings:
                logger.warning(
                    "Runtime MCP connect warning in smart tool call",
                    data={"warning": warning, "agent": default_agent_name},
                )

        if disable_streaming:
            for agent in agents_map.values():
                setter = getattr(agent, "force_non_streaming_next_turn", None)
                if callable(setter):
                    setter(reason="parallel smart tool calls")
            logger.info(
                "Disabled streaming for smart tool child agents",
                data={"agent_count": len(agents_map)},
            )

        await _hydrate_smart_agents_for_execution(agents_map, bundle)

        return await app.send(message)
    finally:
        await _shutdown_agents(agents_map)


async def _run_validate_call(
    context: "Context | None",
    agent_card_path: str,
) -> str:
    resolved_path = _resolve_agent_card_path(agent_card_path, context)

    server_names = None
    if context and context.config and context.config.mcp and context.config.mcp.servers:
        server_names = set(context.config.mcp.servers.keys())

    results = scan_agent_card_path(resolved_path, server_names=server_names)
    return _format_validation_results(results)


async def _run_smart_list_resources_call(
    context: "Context | None",
    agent_card_path: str,
    *,
    server_name: str | None = None,
    mcp_connect: Sequence[str] | None = None,
) -> str:
    resources: Mapping[str, list[str]] = {}
    templates: Mapping[str, Sequence[Any]] = {}
    mcp_server_names: list[str] = []

    if _include_mcp_resources(server_name):
        agents_map, _bundle, default_agent_name, tool_only_agents = await _create_smart_tool_agents(
            context,
            agent_card_path,
        )
        try:
            app = AgentApp(agents_map, tool_only_agents=tool_only_agents)
            if mcp_connect:
                await _apply_runtime_mcp_connections(
                    context=context,
                    agents_map=agents_map,
                    target_agent_name=default_agent_name,
                    mcp_connect=mcp_connect,
                )

            resources = await app.list_resources(
                server_name=server_name,
                agent_name=default_agent_name,
            )

            default_agent = agents_map.get(default_agent_name)
            aggregator = getattr(default_agent, "aggregator", None) if default_agent else None
            list_templates = getattr(aggregator, "list_resource_templates", None)
            if callable(list_templates):
                templates = await list_templates(server_name)

            list_attached = getattr(default_agent, "list_attached_mcp_servers", None)
            if callable(list_attached):
                try:
                    attached = list_attached()
                except Exception:
                    attached = []
                mcp_server_names = sorted(
                    set(attached) | set(resources.keys()) | set(templates.keys())
                )
        finally:
            await _shutdown_agents(agents_map)

    internal_resources: Sequence[InternalResource] = ()
    if _include_internal_resources(server_name):
        internal_resources = list_internal_resources()

    return _format_smart_resource_listing(
        resources,
        templates,
        mcp_server_names=mcp_server_names,
        internal_resources=internal_resources,
    )


async def _run_current_agent_get_resource_call(
    agent: Any,
    resource_uri: str,
    *,
    server_name: str | None = None,
) -> str:
    result = await agent.get_resource(resource_uri=resource_uri, namespace=server_name)
    body = _extract_read_resource_text(result)
    header = f"Resource: {resource_uri}"
    if server_name:
        header += f" (server={server_name})"
    return f"{header}\n\n{body}" if body else header


async def _run_smart_get_resource_call(
    context: "Context | None",
    agent_card_path: str,
    resource_uri: str,
    *,
    server_name: str | None = None,
    mcp_connect: Sequence[str] | None = None,
) -> str:
    agents_map, _bundle, default_agent_name, tool_only_agents = await _create_smart_tool_agents(
        context,
        agent_card_path,
    )
    try:
        app = AgentApp(agents_map, tool_only_agents=tool_only_agents)
        if mcp_connect:
            await _apply_runtime_mcp_connections(
                context=context,
                agents_map=agents_map,
                target_agent_name=default_agent_name,
                mcp_connect=mcp_connect,
            )

        result = await app.get_resource(
            resource_uri=resource_uri,
            server_name=server_name,
            agent_name=default_agent_name,
        )
        body = _extract_read_resource_text(result)
        header = f"Resource: {resource_uri}"
        if server_name:
            header += f" (server={server_name})"
        return f"{header}\n\n{body}" if body else header
    finally:
        await _shutdown_agents(agents_map)


async def _run_smart_with_resource_call(
    context: "Context | None",
    agent_card_path: str,
    message: str,
    resource_uri: str,
    *,
    server_name: str | None = None,
    mcp_connect: Sequence[str] | None = None,
) -> str:
    agents_map, bundle, default_agent_name, tool_only_agents = await _create_smart_tool_agents(
        context,
        agent_card_path,
    )
    try:
        app = AgentApp(agents_map, tool_only_agents=tool_only_agents)
        if mcp_connect:
            await _apply_runtime_mcp_connections(
                context=context,
                agents_map=agents_map,
                target_agent_name=default_agent_name,
                mcp_connect=mcp_connect,
            )

        await _hydrate_smart_agents_for_execution(agents_map, bundle)

        return await app.with_resource(
            prompt_content=message,
            resource_uri=resource_uri,
            server_name=server_name,
            agent_name=default_agent_name,
        )
    finally:
        await _shutdown_agents(agents_map)


def _resolve_template_server_name(
    template_uri: str,
    templates_by_server: Mapping[str, Sequence[Any]],
) -> str | None:
    matches: list[str] = []
    for server_name, templates in templates_by_server.items():
        for template in templates:
            if getattr(template, "uriTemplate", None) == template_uri:
                matches.append(server_name)
                break

    unique_matches = sorted(set(matches))
    if len(unique_matches) == 1:
        return unique_matches[0]
    return None


async def _run_smart_complete_resource_argument_call(
    context: "Context | None",
    agent_card_path: str,
    template_uri: str,
    argument_name: str,
    value: str,
    *,
    server_name: str | None = None,
    context_args: Mapping[str, str] | None = None,
    mcp_connect: Sequence[str] | None = None,
) -> str:
    agents_map, _bundle, default_agent_name, tool_only_agents = await _create_smart_tool_agents(
        context,
        agent_card_path,
    )
    try:
        del tool_only_agents
        if mcp_connect:
            await _apply_runtime_mcp_connections(
                context=context,
                agents_map=agents_map,
                target_agent_name=default_agent_name,
                mcp_connect=mcp_connect,
            )

        default_agent = agents_map.get(default_agent_name)
        aggregator = getattr(default_agent, "aggregator", None) if default_agent else None
        if aggregator is None:
            raise AgentConfigError(
                "Smart resource completion requires MCP-capable card agent",
                f"Agent '{default_agent_name}' does not expose an MCP aggregator",
            )

        target_server = server_name
        if target_server is None:
            templates = await aggregator.list_resource_templates(None)
            target_server = _resolve_template_server_name(template_uri, templates)
            if target_server is None:
                raise AgentConfigError(
                    "Unable to resolve resource template server",
                    "Pass server_name explicitly when multiple servers expose the same template.",
                )

        completion = await aggregator.complete_resource_argument(
            server_name=target_server,
            template_uri=template_uri,
            argument_name=argument_name,
            value=value,
            context_args=dict(context_args) if context_args else None,
        )
        values = completion.values or []
        if not values:
            return "No completion values returned."
        return "\n".join(values)
    finally:
        await _shutdown_agents(agents_map)


def _slash_command_tool_description() -> str:
    return (
        "Execute a fast-agent slash command using native `/...` syntax. "
        "Use `/commands` or `/commands --json` to discover capabilities. "
        "Supports `/skills` (including available/search/help), `/cards`, `/model`, `/mcp`, "
        "`/tools`, `/prompts`, "
        "`/usage`, `/system`, `/markdown`, and `/check`."
    )


def _enable_smart_tooling(agent: Any) -> None:
    """Register smart tool endpoints on a smart-capable agent."""
    setattr(agent, "_parallel_smart_tool_calls", False)
    smart_tool_names = {
        "smart",
        "slash_command",
        "get_resource",
    }
    setattr(agent, "_smart_tool_names", set(smart_tool_names))

    smart_tool = build_default_function_tool(
        agent.smart,
        name="smart",
        description=(
            "Operate on AgentCards. Use action=`run` (default) to load cards from a path and "
            "send a message to the resolved default card agent (default:true, otherwise first "
            "non-tool_only). Use action=`validate` to check AgentCard files without running "
            "them. Optional `mcp_connect` entries accept `/mcp connect` style target strings "
            "for runtime MCP attachment during run calls."
        ),
    )
    slash_command_tool = build_default_function_tool(
        agent.slash_command,
        name="slash_command",
        description=_slash_command_tool_description(),
    )
    resource_read_tool = build_default_function_tool(
        agent.read_resource,
        name="get_resource",
        description=(
            "Read a resource by URI. `internal://` URIs are read directly from bundled "
            "resources shown in the prompt; other URIs are fetched from attached MCP resources "
            "when available."
        ),
    )
    agent.add_tool(smart_tool)
    agent.add_tool(slash_command_tool)
    agent.add_tool(resource_read_tool)


async def _dispatch_smart_tool(
    agent: Any,
    agent_card_path: str,
    message: str | None = None,
    mcp_connect: list[str] | None = None,
    action: Literal["run", "validate"] = "run",
) -> str:
    context = getattr(agent, "context", None)
    if action == "validate":
        return await _run_validate_call(context, agent_card_path)
    if action != "run":
        raise AgentConfigError(
            "Invalid smart action",
            "Supported smart actions: run, validate.",
        )

    message_text = (message or "").strip()
    if not message_text:
        raise AgentConfigError(
            "Missing smart message",
            "Provide `message` when action=`run`.",
        )

    disable_streaming = bool(getattr(agent, "_parallel_smart_tool_calls", False))
    return await _run_smart_call(
        context,
        agent_card_path,
        message_text,
        mcp_connect=mcp_connect,
        disable_streaming=disable_streaming,
    )


async def _dispatch_slash_command_tool(agent: Any, command: str) -> str:
    return await _run_slash_command_call(agent, command)


async def _dispatch_resource_read_tool(
    agent: Any,
    uri: str,
    server_name: str | None = None,
) -> str:
    if _is_internal_resource_uri(uri):
        return await _run_internal_resource_read_call(uri)
    return await _run_current_agent_get_resource_call(
        agent,
        uri,
        server_name=server_name,
    )


async def _dispatch_smart_list_resources_tool(
    agent: Any,
    agent_card_path: str,
    mcp_connect: list[str] | None = None,
    server_name: str | None = None,
) -> str:
    context = getattr(agent, "context", None)
    return await _run_smart_list_resources_call(
        context,
        agent_card_path,
        server_name=server_name,
        mcp_connect=mcp_connect,
    )


async def _dispatch_smart_get_resource_tool(
    agent: Any,
    agent_card_path: str,
    resource_uri: str,
    server_name: str | None = None,
    mcp_connect: list[str] | None = None,
) -> str:
    if _is_internal_resource_uri(resource_uri):
        return await _run_internal_resource_read_call(resource_uri)

    context = getattr(agent, "context", None)
    return await _run_smart_get_resource_call(
        context,
        agent_card_path,
        resource_uri,
        server_name=server_name,
        mcp_connect=mcp_connect,
    )


async def _dispatch_smart_with_resource_tool(
    agent: Any,
    agent_card_path: str,
    message: str,
    resource_uri: str,
    server_name: str | None = None,
    mcp_connect: list[str] | None = None,
) -> str:
    context = getattr(agent, "context", None)
    return await _run_smart_with_resource_call(
        context,
        agent_card_path,
        message,
        resource_uri,
        server_name=server_name,
        mcp_connect=mcp_connect,
    )


async def _dispatch_smart_complete_resource_argument_tool(
    agent: Any,
    agent_card_path: str,
    template_uri: str,
    argument_name: str,
    value: str,
    server_name: str | None = None,
    context_args: dict[str, str] | None = None,
    mcp_connect: list[str] | None = None,
) -> str:
    context = getattr(agent, "context", None)
    return await _run_smart_complete_resource_argument_call(
        context,
        agent_card_path,
        template_uri,
        argument_name,
        value,
        server_name=server_name,
        context_args=context_args,
        mcp_connect=mcp_connect,
    )


class SmartAgent(McpAgent):
    """Smart agent with built-in tools for AgentCard execution and validation."""

    def __init__(
        self,
        config: AgentConfig,
        context: Context | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(config=config, context=context, **kwargs)
        _enable_smart_tooling(self)

    @property
    def agent_type(self) -> AgentType:
        return AgentType.SMART

    async def smart(
        self,
        agent_card_path: str,
        message: str | None = None,
        mcp_connect: list[str] | None = None,
        action: Literal["run", "validate"] = "run",
    ) -> str:
        """Run or validate AgentCards."""
        return await _dispatch_smart_tool(
            self,
            agent_card_path,
            message,
            mcp_connect=mcp_connect,
            action=action,
        )

    async def slash_command(self, command: str) -> str:
        """Execute a slash command using `/...` syntax."""
        return await _dispatch_slash_command_tool(self, command)

    async def read_resource(self, uri: str, server_name: str | None = None) -> str:
        """Read a bundled internal resource or attached MCP resource by URI."""
        return await _dispatch_resource_read_tool(self, uri, server_name=server_name)

    async def smart_list_resources(
        self,
        agent_card_path: str,
        mcp_connect: list[str] | None = None,
        server_name: str | None = None,
    ) -> str:
        """List card-agent resources and bundled internal resources."""
        return await _dispatch_smart_list_resources_tool(
            self,
            agent_card_path,
            mcp_connect=mcp_connect,
            server_name=server_name,
        )

    async def smart_get_resource(
        self,
        agent_card_path: str,
        resource_uri: str,
        server_name: str | None = None,
        mcp_connect: list[str] | None = None,
    ) -> str:
        """Read a card-agent resource or bundled internal resource by URI."""
        return await _dispatch_smart_get_resource_tool(
            self,
            agent_card_path,
            resource_uri,
            server_name=server_name,
            mcp_connect=mcp_connect,
        )

    async def smart_with_resource(
        self,
        agent_card_path: str,
        message: str,
        resource_uri: str,
        server_name: str | None = None,
        mcp_connect: list[str] | None = None,
    ) -> str:
        """Run a card agent prompt with a resource attachment."""
        return await _dispatch_smart_with_resource_tool(
            self,
            agent_card_path,
            message,
            resource_uri,
            server_name=server_name,
            mcp_connect=mcp_connect,
        )

    async def smart_complete_resource_argument(
        self,
        agent_card_path: str,
        template_uri: str,
        argument_name: str,
        value: str,
        server_name: str | None = None,
        context_args: dict[str, str] | None = None,
        mcp_connect: list[str] | None = None,
    ) -> str:
        """Run MCP completion for a resource-template argument."""
        return await _dispatch_smart_complete_resource_argument_tool(
            self,
            agent_card_path,
            template_uri,
            argument_name,
            value,
            server_name=server_name,
            context_args=context_args,
            mcp_connect=mcp_connect,
        )


class SmartAgentsAsToolsAgent(AgentsAsToolsAgent):
    """Agents-as-tools wrapper with smart tools."""

    def __init__(
        self,
        config: AgentConfig,
        agents: list["LlmAgent"],
        options: AgentsAsToolsOptions | None = None,
        context: Any | None = None,
        child_message_files: dict[str, list[Path]] | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            config=config,
            agents=agents,
            options=options,
            context=context,
            child_message_files=child_message_files,
            **kwargs,
        )
        _enable_smart_tooling(self)

    @property
    def agent_type(self) -> AgentType:
        return AgentType.SMART

    async def smart(
        self,
        agent_card_path: str,
        message: str | None = None,
        mcp_connect: list[str] | None = None,
        action: Literal["run", "validate"] = "run",
    ) -> str:
        return await _dispatch_smart_tool(
            self,
            agent_card_path,
            message,
            mcp_connect=mcp_connect,
            action=action,
        )

    async def slash_command(self, command: str) -> str:
        return await _dispatch_slash_command_tool(self, command)

    async def read_resource(self, uri: str, server_name: str | None = None) -> str:
        return await _dispatch_resource_read_tool(self, uri, server_name=server_name)

    async def smart_list_resources(
        self,
        agent_card_path: str,
        mcp_connect: list[str] | None = None,
        server_name: str | None = None,
    ) -> str:
        return await _dispatch_smart_list_resources_tool(
            self,
            agent_card_path,
            mcp_connect=mcp_connect,
            server_name=server_name,
        )

    async def smart_get_resource(
        self,
        agent_card_path: str,
        resource_uri: str,
        server_name: str | None = None,
        mcp_connect: list[str] | None = None,
    ) -> str:
        return await _dispatch_smart_get_resource_tool(
            self,
            agent_card_path,
            resource_uri,
            server_name=server_name,
            mcp_connect=mcp_connect,
        )

    async def smart_with_resource(
        self,
        agent_card_path: str,
        message: str,
        resource_uri: str,
        server_name: str | None = None,
        mcp_connect: list[str] | None = None,
    ) -> str:
        return await _dispatch_smart_with_resource_tool(
            self,
            agent_card_path,
            message,
            resource_uri,
            server_name=server_name,
            mcp_connect=mcp_connect,
        )

    async def smart_complete_resource_argument(
        self,
        agent_card_path: str,
        template_uri: str,
        argument_name: str,
        value: str,
        server_name: str | None = None,
        context_args: dict[str, str] | None = None,
        mcp_connect: list[str] | None = None,
    ) -> str:
        return await _dispatch_smart_complete_resource_argument_tool(
            self,
            agent_card_path,
            template_uri,
            argument_name,
            value,
            server_name=server_name,
            context_args=context_args,
            mcp_connect=mcp_connect,
        )


class SmartAgentWithUI(McpUIMixin, SmartAgent):
    """Smart agent with UI support."""

    def __init__(
        self,
        config: AgentConfig,
        context: Context | None = None,
        ui_mode: str = "auto",
        **kwargs: Any,
    ) -> None:
        super().__init__(config=config, context=context, ui_mode=ui_mode, **kwargs)



"""
Direct factory functions for creating agent and workflow instances without proxies.
Implements type-safe factories with improved error handling.
"""

from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import Any, Awaitable, Callable, Mapping, Protocol, Sequence, TypeVar, cast

from fastmcp.tools import FunctionTool

from fast_agent.agents import McpAgent
from fast_agent.agents.agent_types import AgentConfig, AgentType
from fast_agent.agents.llm_agent import LlmAgent
from fast_agent.agents.workflow.evaluator_optimizer import (
    EvaluatorOptimizerAgent,
    QualityRating,
)
from fast_agent.agents.workflow.iterative_planner import IterativePlanner
from fast_agent.agents.workflow.parallel_agent import ParallelAgent
from fast_agent.agents.workflow.router_agent import RouterAgent
from fast_agent.context import Context
from fast_agent.core import Core
from fast_agent.core.agent_card_types import AgentCardData
from fast_agent.core.exceptions import AgentConfigError, ModelConfigError
from fast_agent.core.logging.logger import get_logger
from fast_agent.core.model_resolution import HARDCODED_DEFAULT_MODEL, resolve_model_spec
from fast_agent.core.validation import (
    get_dependencies_groups,
    is_basic_like_agent_type,
    normalize_agent_type_value,
)
from fast_agent.event_progress import ProgressAction
from fast_agent.hooks.hook_messages import show_hook_failure
from fast_agent.interfaces import (
    AgentProtocol,
    LLMFactoryProtocol,
    ModelFactoryFunctionProtocol,
)
from fast_agent.llm.model_factory import ModelFactory
from fast_agent.mcp.ui_agent import McpAgentWithUI
from fast_agent.tools.function_tool_loader import load_function_tools
from fast_agent.tools.hook_loader import load_tool_runner_hooks
from fast_agent.types import RequestParams

# Type aliases for improved readability and IDE support
AgentDict = dict[str, AgentProtocol]
AgentConfigDict = Mapping[str, AgentCardData | dict[str, Any]]
AgentTypeBuilder = Callable[[str, Mapping[str, Any], "AgentBuildContext", AgentDict], Awaitable[None]]


@dataclass(frozen=True)
class AgentBuildContext:
    app_instance: "CoreContextProtocol"
    agents_dict: AgentConfigDict
    active_agents: AgentDict
    model_factory_func: ModelFactoryFunctionProtocol
    session_history_enabled: bool


@dataclass(frozen=True)
class AgentsAsToolsBuildInputs:
    config: AgentConfig
    function_tools: list[FunctionTool]
    child_agents: list[AgentProtocol]
    options: Any
    child_message_files: dict[str, list[Path]]


class CoreContextProtocol(Protocol):
    context: Context


class _ContextCoreShim:
    def __init__(self, context: Context) -> None:
        self.context = context


def _ensure_basic_only_agents(agents_dict: AgentConfigDict) -> None:
    for name, agent_data in agents_dict.items():
        agent_type = agent_data.get("type") if isinstance(agent_data, Mapping) else None
        if not is_basic_like_agent_type(agent_type):
            raise AgentConfigError(
                "Smart tool only supports 'agent' cards",
                f"Card '{name}' has unsupported type '{agent_type}'",
            )

def _load_configured_function_tools(
    config: AgentConfig,
    agent_data: Mapping[str, Any],
) -> list[FunctionTool]:
    """Load function tools from config or agent_data, resolving paths.

    Args:
        config: The agent configuration
        agent_data: The agent data dictionary from card or registry

    Returns:
        List of loaded function tools (may be empty)
    """
    tools_config_raw = config.function_tools
    if tools_config_raw is None:
        tools_config_raw = agent_data.get("function_tools")

    tools_config: list[Callable[..., Any] | str] | None = None
    if isinstance(tools_config_raw, str):
        tools_config = [tools_config_raw]
    elif isinstance(tools_config_raw, list):
        tools_config = cast("list[Callable[..., Any] | str]", tools_config_raw)

    if not tools_config:
        return []

    source_path = agent_data.get("source_path")
    base_path = Path(source_path).parent if source_path else None
    return load_function_tools(tools_config, base_path)


def _register_loaded_agent(
    result_agents: AgentDict,
    name: str,
    agent: AgentProtocol,
) -> None:
    result_agents[name] = agent
    logger.info(
        f"Loaded {name}",
        data={
            "progress_action": ProgressAction.LOADED,
            "agent_name": name,
            "target": name,
        },
    )


def _iter_agents_of_type(
    agents_dict: AgentConfigDict,
    agent_type: AgentType,
) -> list[tuple[str, Mapping[str, Any]]]:
    return [
        (name, agent_data)
        for name, agent_data in agents_dict.items()
        if normalize_agent_type_value(agent_data.get("type")) == agent_type.value
    ]


def _resolve_child_agents(
    parent_name: str,
    child_names: Sequence[str],
    active_agents: AgentDict,
    *,
    skip_missing: bool,
) -> list[AgentProtocol]:
    child_agents: list[AgentProtocol] = []
    for agent_name in child_names:
        child_agent = active_agents.get(agent_name)
        if child_agent is None:
            if skip_missing:
                logger.warning(
                    "Skipping missing child agent",
                    data={"agent_name": agent_name, "parent": parent_name},
                )
                continue
            raise AgentConfigError(f"Agent {agent_name} not found")
        child_agents.append(child_agent)
    return child_agents


def _build_agents_as_tools_options(agent_data: Mapping[str, Any]) -> object:
    from fast_agent.agents.workflow.agents_as_tools_agent import AgentsAsToolsOptions

    raw_opts = agent_data.get("agents_as_tools_options") or {}
    opt_kwargs = {key: value for key, value in raw_opts.items() if value is not None}
    return AgentsAsToolsOptions(**opt_kwargs)


def _collect_child_message_files(
    child_names: Sequence[str],
    agents_dict: AgentConfigDict,
    options: object,
) -> dict[str, list[Path]]:
    from fast_agent.agents.workflow.agents_as_tools_agent import (
        HistoryMergeTarget,
        HistorySource,
    )

    child_message_files: dict[str, list[Path]] = {}
    history_source = getattr(options, "history_source", None)
    history_merge_target = getattr(options, "history_merge_target", None)
    requires_messages = (
        history_source == HistorySource.MESSAGES
        or history_merge_target == HistoryMergeTarget.MESSAGES
    )
    missing_messages: list[str] = []

    for agent_name in child_names:
        child_data = agents_dict.get(agent_name, {})
        message_files = child_data.get("message_files")
        if not message_files:
            if requires_messages:
                missing_messages.append(agent_name)
            continue
        child_message_files[agent_name] = list(message_files)

    if missing_messages:
        missing_list = ", ".join(sorted(set(missing_messages)))
        raise AgentConfigError(
            "history_source/history_merge_target=messages requires child agents with messages",
            f"Missing messages for: {missing_list}",
        )

    return child_message_files


def _build_agents_as_tools_inputs(
    name: str,
    agent_data: Mapping[str, Any],
    build_ctx: AgentBuildContext,
) -> AgentsAsToolsBuildInputs:
    config = cast("AgentConfig", agent_data["config"])
    child_names = cast("Sequence[str]", agent_data.get("child_agents", []) or [])
    options = _build_agents_as_tools_options(agent_data)
    return AgentsAsToolsBuildInputs(
        config=config,
        function_tools=_load_configured_function_tools(config, agent_data),
        child_agents=_resolve_child_agents(
            name,
            child_names,
            build_ctx.active_agents,
            skip_missing=True,
        ),
        options=options,
        child_message_files=_collect_child_message_files(
            child_names,
            build_ctx.agents_dict,
            options,
        ),
    )


async def _initialize_agent_with_llm(
    agent: Any,
    config: AgentConfig,
    model_factory_func: ModelFactoryFunctionProtocol,
) -> None:
    await agent.initialize()
    llm_factory = model_factory_func(model=config.model)
    await agent.attach_llm(
        llm_factory,
        request_params=config.default_request_params,
        api_key=config.api_key,
    )


def _attach_child_agents_as_tools(
    agent: Any,
    parent_name: str,
    child_names: Sequence[str],
    active_agents: AgentDict,
) -> None:
    add_tool_fn = getattr(agent, "add_agent_tool", None)
    if not callable(add_tool_fn):
        return

    for child_agent in _resolve_child_agents(
        parent_name,
        child_names,
        active_agents,
        skip_missing=True,
    ):
        add_tool_fn(child_agent)




async def _finalize_agent(
    agent: LlmAgent,
    name: str,
    config: AgentConfig,
    agent_data: Mapping[str, Any],
    model_factory_func: ModelFactoryFunctionProtocol,
    result_agents: AgentDict,
    session_history_enabled: bool,
) -> None:
    """Complete agent setup: initialize, attach LLM, apply hooks, register.

    Args:
        agent: The agent instance to finalize
        name: Agent name for registration
        config: Agent configuration
        agent_data: Agent data dictionary (for source_path)
        model_factory_func: Factory function for LLM creation
        result_agents: Dictionary to register the agent in
        session_history_enabled: Whether session history tracking is enabled
    """
    await agent.initialize()

    llm_factory = model_factory_func(model=config.model)
    await agent.attach_llm(
        llm_factory,
        request_params=config.default_request_params,
        api_key=config.api_key,
    )

    if config.tool_hooks or config.trim_tool_history or session_history_enabled:
        _apply_tool_hooks(
            agent,
            config,
            agent_data.get("source_path"),
            enable_session_history=session_history_enabled,
        )

    _register_loaded_agent(result_agents, name, agent)


T = TypeVar("T")  # For generic types


logger = get_logger(__name__)


def _create_agent_with_ui_if_needed(
    agent_class: type,
    config: Any,
    context: Any,
    **kwargs: Any,
) -> Any:
    """
    Create an agent with UI support if MCP UI mode is enabled.

    Args:
        agent_class: The agent class to potentially enhance with UI
        config: Agent configuration
        context: Application context
        **kwargs: Additional arguments passed to agent constructor (e.g., tools)

    Returns:
        Either a UI-enhanced agent instance or the original agent instance
    """
    # Check UI mode from settings
    settings = context.config if hasattr(context, "config") else None
    ui_mode = getattr(settings, "mcp_ui_mode", "auto") if settings else "auto"

    if ui_mode != "disabled" and agent_class == McpAgent:
        # Use the UI-enhanced agent class instead of the base class
        return McpAgentWithUI(config=config, context=context, ui_mode=ui_mode, **kwargs)
    else:
        # Create the original agent instance
        return agent_class(config=config, context=context, **kwargs)


def _apply_tool_hooks(
    agent: LlmAgent,
    config: AgentConfig,
    source_path: str | None = None,
    *,
    enable_session_history: bool = False,
) -> None:
    """
    Apply tool runner hooks to an agent based on config.

    Handles both:
    - tool_hooks: dict mapping hook types to function specs
    - trim_tool_history: shortcut to apply built-in history trimmer

    Args:
        agent: The agent to apply hooks to
        config: Agent configuration with tool_hooks and/or trim_tool_history
        source_path: Path to the source file for resolving relative hook paths
    """
    # Import here to avoid circular imports
    from fast_agent.agents.tool_runner import ToolRunnerHooks
    from fast_agent.hooks import save_session_history, trim_tool_loop_history
    from fast_agent.hooks.hook_context import HookContext

    hooks_config = config.tool_hooks
    trim_history = config.trim_tool_history

    # If trim_tool_history is set and no after_turn_complete hook is configured,
    # add the built-in trimmer
    if trim_history:
        if hooks_config is None:
            hooks_config = {}
        if "after_turn_complete" not in hooks_config:
            # Use a wrapper that creates HookContext for the built-in trimmer
            async def _trimmer_wrapper(runner, message):
                ctx = HookContext(
                    runner=runner,
                    agent=agent,
                    message=message,
                    hook_type="after_turn_complete",
                )
                try:
                    await trim_tool_loop_history(ctx)
                except Exception as exc:  # noqa: BLE001
                    show_hook_failure(
                        ctx,
                        hook_name="trim_tool_loop_history",
                        hook_kind="tool",
                        error=exc,
                    )
                    logger.exception(
                        "Tool hook failed",
                        hook_type="after_turn_complete",
                        hook_name="trim_tool_loop_history",
                    )
                    raise

            # Set the hooks directly since we have a callable, not a spec string
            existing_hooks = getattr(agent, "tool_runner_hooks", None) or ToolRunnerHooks()
            agent.tool_runner_hooks = ToolRunnerHooks(
                before_llm_call=existing_hooks.before_llm_call,
                after_llm_call=existing_hooks.after_llm_call,
                before_tool_call=existing_hooks.before_tool_call,
                after_tool_call=existing_hooks.after_tool_call,
                after_turn_complete=_trimmer_wrapper,
            )
            # Keep going so session history hooks can still be applied

    # Load custom hooks from config
    if hooks_config:
        base_path = Path(source_path).parent if source_path else None
        loaded_hooks = load_tool_runner_hooks(hooks_config, agent, base_path)
        if loaded_hooks:
            # Merge with any existing hooks (trim_tool_history wrapper)
            existing = getattr(agent, "tool_runner_hooks", None)
            if existing:
                agent.tool_runner_hooks = ToolRunnerHooks(
                    before_llm_call=loaded_hooks.before_llm_call or existing.before_llm_call,
                    after_llm_call=loaded_hooks.after_llm_call or existing.after_llm_call,
                    before_tool_call=loaded_hooks.before_tool_call or existing.before_tool_call,
                    after_tool_call=loaded_hooks.after_tool_call or existing.after_tool_call,
                    after_turn_complete=loaded_hooks.after_turn_complete or existing.after_turn_complete,
                )
            else:
                agent.tool_runner_hooks = loaded_hooks

    if enable_session_history:
        existing_hooks = getattr(agent, "tool_runner_hooks", None) or ToolRunnerHooks()
        existing_after_turn = existing_hooks.after_turn_complete

        async def _session_history_wrapper(runner, message):
            if existing_after_turn is not None:
                await existing_after_turn(runner, message)
            ctx = HookContext(
                runner=runner,
                agent=agent,
                message=message,
                hook_type="after_turn_complete",
            )
            try:
                await save_session_history(ctx)
            except Exception as exc:  # noqa: BLE001
                show_hook_failure(
                    ctx,
                    hook_name="save_session_history",
                    hook_kind="tool",
                    error=exc,
                )
                logger.exception(
                    "Tool hook failed",
                    hook_type="after_turn_complete",
                    hook_name="save_session_history",
                )
                raise

        agent.tool_runner_hooks = ToolRunnerHooks(
            before_llm_call=existing_hooks.before_llm_call,
            after_llm_call=existing_hooks.after_llm_call,
            before_tool_call=existing_hooks.before_tool_call,
            after_tool_call=existing_hooks.after_tool_call,
            after_turn_complete=_session_history_wrapper,
        )


class AgentCreatorProtocol(Protocol):
    """Protocol for agent creator functions."""

    async def __call__(
        self,
        app_instance: Core,
        agents_dict: AgentConfigDict,
        agent_type: AgentType,
        active_agents: AgentDict | None = None,
        model_factory_func: ModelFactoryFunctionProtocol | None = None,
        **kwargs: Any,
    ) -> AgentDict: ...



def get_model_factory(
    context,
    model: str | None = None,
    request_params: RequestParams | None = None,
    default_model: str | None = None,
    cli_model: str | None = None,
) -> LLMFactoryProtocol:
    """
    Get model factory using specified or default model.
    Model string is parsed by ModelFactory to determine provider and reasoning effort.

    Precedence (lowest to highest):
        1. Hardcoded default (gpt-5-mini?reasoning=low)
        2. FAST_AGENT_MODEL environment variable
        3. Config file default_model
        4. CLI --model argument
        5. Decorator model parameter

    Args:
        context: Application context
        model: Optional model specification string (highest precedence)
        request_params: Optional RequestParams to configure LLM behavior
        default_model: Default model from configuration
        cli_model: Model specified via command line

    Returns:
        ModelFactory instance for the specified or default model
    """
    model_spec, source = resolve_model_spec(
        context,
        model=model,
        default_model=default_model,
        cli_model=cli_model,
        hardcoded_default=HARDCODED_DEFAULT_MODEL,
    )
    if model_spec is None:
        raise ModelConfigError(
            "No model configured",
            "Set --model, FAST_AGENT_MODEL, or default_model in config.",
        )
    logger.info(
        f"Resolved model '{model_spec}' via {source}",
        model=model_spec,
        source=source,
    )

    # Update or create request_params with the final model choice
    if request_params:
        request_params = request_params.model_copy(update={"model": model_spec})
    else:
        request_params = RequestParams(model=model_spec)

    # Let model factory handle the model string parsing and setup
    return ModelFactory.create_factory(model_spec)


def get_default_model_source(
    config_default_model: str | None = None,
    cli_model: str | None = None,
    model_references: dict[str, dict[str, str]] | None = None,
) -> str | None:
    """
    Determine the source of the default model selection.
    Returns "environment variable", "config file", or None (if CLI or hardcoded default).

    This is used to display informational messages about where the model
    configuration is coming from. Only shows a message for env var or config file,
    not for explicit CLI usage or the hardcoded system default.
    """
    # CLI model is explicit - no message needed
    if cli_model:
        return None

    _, source = resolve_model_spec(
        context=None,
        default_model=config_default_model,
        cli_model=None,
        fallback_to_hardcoded=False,
        model_references=model_references,
    )
    if source == "config file":
        return "config file"
    if source and source.startswith("environment variable"):
        return "environment variable"

    return None


async def _create_basic_agent(
    name: str,
    agent_data: Mapping[str, Any],
    build_ctx: AgentBuildContext,
    result_agents: AgentDict,
) -> None:
    config = cast("AgentConfig", agent_data["config"])
    child_names = cast("Sequence[str]", agent_data.get("child_agents", []) or [])
    if child_names:
        inputs = _build_agents_as_tools_inputs(name, agent_data, build_ctx)

        from fast_agent.agents.workflow.agents_as_tools_agent import AgentsAsToolsAgent

        agent = AgentsAsToolsAgent(
            config=inputs.config,
            context=build_ctx.app_instance.context,
            agents=cast("list[LlmAgent]", inputs.child_agents),
            options=inputs.options,
            tools=inputs.function_tools,
            child_message_files=inputs.child_message_files,
        )
    else:
        function_tools = _load_configured_function_tools(config, agent_data)
        agent = _create_agent_with_ui_if_needed(
            McpAgent,
            config,
            build_ctx.app_instance.context,
            tools=function_tools,
        )

    await _finalize_agent(
        agent,
        name,
        config,
        agent_data,
        build_ctx.model_factory_func,
        result_agents,
        build_ctx.session_history_enabled,
    )


async def _create_smart_agent(
    name: str,
    agent_data: Mapping[str, Any],
    build_ctx: AgentBuildContext,
    result_agents: AgentDict,
) -> None:
    config = cast("AgentConfig", agent_data["config"])
    child_names = cast("Sequence[str]", agent_data.get("child_agents", []) or [])
    if child_names:
        inputs = _build_agents_as_tools_inputs(name, agent_data, build_ctx)

        from fast_agent.agents.smart_agent import SmartAgentsAsToolsAgent

        agent = SmartAgentsAsToolsAgent(
            config=inputs.config,
            context=build_ctx.app_instance.context,
            agents=cast("list[LlmAgent]", inputs.child_agents),
            options=inputs.options,
            tools=inputs.function_tools,
            child_message_files=inputs.child_message_files,
        )
    else:
        function_tools = _load_configured_function_tools(config, agent_data)

        from fast_agent.agents.smart_agent import SmartAgent, SmartAgentWithUI

        settings = (
            build_ctx.app_instance.context.config if build_ctx.app_instance.context else None
        )
        ui_mode = getattr(settings, "mcp_ui_mode", "auto") if settings else "auto"
        if ui_mode != "disabled":
            agent = SmartAgentWithUI(
                config=config,
                context=build_ctx.app_instance.context,
                ui_mode=ui_mode,
                tools=function_tools,
            )
        else:
            agent = SmartAgent(
                config=config,
                context=build_ctx.app_instance.context,
                tools=function_tools,
            )

    await _finalize_agent(
        agent,
        name,
        config,
        agent_data,
        build_ctx.model_factory_func,
        result_agents,
        build_ctx.session_history_enabled,
    )


async def _create_custom_agent(
    name: str,
    agent_data: Mapping[str, Any],
    build_ctx: AgentBuildContext,
    result_agents: AgentDict,
) -> None:
    config = cast("AgentConfig", agent_data["config"])
    cls = agent_data.get("agent_class") or agent_data.get("cls")
    if cls is None:
        raise AgentConfigError(
            f"Custom agent '{name}' missing class reference ('agent_class' or 'cls')"
        )

    agent = _create_agent_with_ui_if_needed(
        cls,
        config,
        build_ctx.app_instance.context,
    )
    await _initialize_agent_with_llm(agent, config, build_ctx.model_factory_func)

    child_names = cast("Sequence[str]", agent_data.get("child_agents", []) or [])
    if child_names:
        _attach_child_agents_as_tools(agent, name, child_names, build_ctx.active_agents)

    if (
        config.tool_hooks
        or config.trim_tool_history
        or build_ctx.session_history_enabled
    ):
        _apply_tool_hooks(
            agent,
            config,
            agent_data.get("source_path"),
            enable_session_history=build_ctx.session_history_enabled,
        )

    _register_loaded_agent(result_agents, name, agent)


async def _create_planner_agent(
    name: str,
    agent_data: Mapping[str, Any],
    build_ctx: AgentBuildContext,
    result_agents: AgentDict,
) -> None:
    config = cast("AgentConfig", agent_data["config"])
    child_names = cast("Sequence[str]", agent_data["child_agents"])
    child_agents = _resolve_child_agents(
        name,
        child_names,
        build_ctx.active_agents,
        skip_missing=False,
    )

    orchestrator = IterativePlanner(
        config=config,
        context=build_ctx.app_instance.context,
        agents=child_agents,
        plan_iterations=agent_data.get("plan_iterations", 5),
        plan_type=agent_data.get("plan_type", "full"),
    )
    await _initialize_agent_with_llm(orchestrator, config, build_ctx.model_factory_func)
    result_agents[name] = orchestrator


async def _create_parallel_workflow_agent(
    name: str,
    agent_data: Mapping[str, Any],
    build_ctx: AgentBuildContext,
    result_agents: AgentDict,
) -> None:
    config = cast("AgentConfig", agent_data["config"])
    fan_in_name = agent_data.get("fan_in")
    fan_out_names = cast("Sequence[str]", agent_data["fan_out"])

    if not fan_in_name:
        fan_in_name = f"{name}_fan_in"
        fan_in_agent = await _create_default_fan_in_agent(
            fan_in_name,
            build_ctx.app_instance.context,
            build_ctx.model_factory_func,
        )
        result_agents[fan_in_name] = fan_in_agent
    else:
        fan_in_agent = build_ctx.active_agents.get(fan_in_name)
        if fan_in_agent is None:
            raise AgentConfigError(f"Fan-in agent {fan_in_name} not found")

    fan_out_agents = _resolve_child_agents(
        name,
        fan_out_names,
        build_ctx.active_agents,
        skip_missing=False,
    )

    parallel = ParallelAgent(
        config=config,
        context=build_ctx.app_instance.context,
        fan_in_agent=fan_in_agent,
        fan_out_agents=fan_out_agents,
        include_request=agent_data.get("include_request", True),
    )
    await parallel.initialize()
    result_agents[name] = parallel


async def _create_router_workflow_agent(
    name: str,
    agent_data: Mapping[str, Any],
    build_ctx: AgentBuildContext,
    result_agents: AgentDict,
) -> None:
    config = cast("AgentConfig", agent_data["config"])
    router = RouterAgent(
        config=config,
        context=build_ctx.app_instance.context,
        agents=cast(
            "list[LlmAgent]",
            _resolve_child_agents(
                name,
                cast("Sequence[str]", agent_data["router_agents"]),
                build_ctx.active_agents,
                skip_missing=False,
            ),
        ),
        routing_instruction=agent_data.get("instruction"),
    )
    await _initialize_agent_with_llm(router, config, build_ctx.model_factory_func)
    result_agents[name] = router


async def _create_chain_workflow_agent(
    name: str,
    agent_data: Mapping[str, Any],
    build_ctx: AgentBuildContext,
    result_agents: AgentDict,
) -> None:
    from fast_agent.agents.workflow.chain_agent import ChainAgent

    config = cast("AgentConfig", agent_data["config"])
    agent_names = cast("Sequence[str]", agent_data["sequence"])
    if not agent_names:
        raise AgentConfigError("No agents in the chain")

    chain = ChainAgent(
        config=config,
        context=build_ctx.app_instance.context,
        agents=cast(
            "list[LlmAgent]",
            _resolve_child_agents(
                name,
                agent_names,
                build_ctx.active_agents,
                skip_missing=False,
            ),
        ),
        cumulative=agent_data.get("cumulative", False),
    )
    await chain.initialize()
    result_agents[name] = chain


async def _create_evaluator_optimizer_agent(
    name: str,
    agent_data: Mapping[str, Any],
    build_ctx: AgentBuildContext,
    result_agents: AgentDict,
) -> None:
    config = cast("AgentConfig", agent_data["config"])
    generator_name = cast("str", agent_data["generator"])
    evaluator_name = cast("str", agent_data["evaluator"])

    generator_agent = build_ctx.active_agents.get(generator_name)
    if generator_agent is None:
        raise AgentConfigError(f"Generator agent {generator_name} not found")

    evaluator_agent = build_ctx.active_agents.get(evaluator_name)
    if evaluator_agent is None:
        raise AgentConfigError(f"Evaluator agent {evaluator_name} not found")

    evaluator_optimizer = EvaluatorOptimizerAgent(
        config=config,
        context=build_ctx.app_instance.context,
        generator_agent=generator_agent,
        evaluator_agent=evaluator_agent,
        min_rating=QualityRating(agent_data.get("min_rating", "GOOD")),
        max_refinements=agent_data.get("max_refinements", 3),
        refinement_instruction=agent_data.get("refinement_instruction"),
    )
    await evaluator_optimizer.initialize()
    result_agents[name] = evaluator_optimizer


async def _create_maker_agent(
    name: str,
    agent_data: Mapping[str, Any],
    build_ctx: AgentBuildContext,
    result_agents: AgentDict,
) -> None:
    from fast_agent.agents.workflow.maker_agent import MakerAgent, MatchStrategy

    config = cast("AgentConfig", agent_data["config"])
    worker_name = cast("str", agent_data["worker"])
    worker_agent = build_ctx.active_agents.get(worker_name)
    if worker_agent is None:
        raise AgentConfigError(f"Worker agent {worker_name} not found")

    maker_agent = MakerAgent(
        config=config,
        context=build_ctx.app_instance.context,
        worker_agent=worker_agent,
        k=agent_data.get("k", 3),
        max_samples=agent_data.get("max_samples", 50),
        match_strategy=MatchStrategy(agent_data.get("match_strategy", "exact")),
        red_flag_max_length=agent_data.get("red_flag_max_length"),
    )
    await maker_agent.initialize()
    result_agents[name] = maker_agent


_AGENT_TYPE_BUILDERS: dict[AgentType, AgentTypeBuilder] = {
    AgentType.LLM: _create_basic_agent,
    AgentType.BASIC: _create_basic_agent,
    AgentType.SMART: _create_smart_agent,
    AgentType.CUSTOM: _create_custom_agent,
    AgentType.ORCHESTRATOR: _create_planner_agent,
    AgentType.ITERATIVE_PLANNER: _create_planner_agent,
    AgentType.PARALLEL: _create_parallel_workflow_agent,
    AgentType.ROUTER: _create_router_workflow_agent,
    AgentType.CHAIN: _create_chain_workflow_agent,
    AgentType.EVALUATOR_OPTIMIZER: _create_evaluator_optimizer_agent,
    AgentType.MAKER: _create_maker_agent,
}


async def create_agents_by_type(
    app_instance: CoreContextProtocol,
    agents_dict: AgentConfigDict,
    agent_type: AgentType,
    model_factory_func: ModelFactoryFunctionProtocol,
    active_agents: AgentDict | None = None,
    **kwargs: Any,
) -> AgentDict:
    """
    Generic method to create agents of a specific type without using proxies.

    Args:
        app_instance: The main application instance
        agents_dict: Dictionary of agent configurations
        agent_type: Type of agents to create
        active_agents: Dictionary of already created agents (for dependencies)
        model_factory_func: Function for creating model factories
        **kwargs: Additional type-specific parameters

    Returns:
        Dictionary of initialized agent instances
    """
    if active_agents is None:
        active_agents = {}

    result_agents: AgentDict = {}
    session_history_enabled = True
    if app_instance.context and app_instance.context.config:
        session_history_enabled = getattr(app_instance.context.config, "session_history", True)
    builder = _AGENT_TYPE_BUILDERS.get(agent_type)
    if builder is None:
        raise ValueError(f"Unknown agent type: {agent_type}")

    build_ctx = AgentBuildContext(
        app_instance=app_instance,
        agents_dict=agents_dict,
        active_agents=active_agents,
        model_factory_func=model_factory_func,
        session_history_enabled=session_history_enabled,
    )

    for name, agent_data in _iter_agents_of_type(agents_dict, agent_type):
        await builder(name, agent_data, build_ctx, result_agents)

    return result_agents


async def active_agents_in_dependency_group(
    app_instance: CoreContextProtocol,
    agents_dict: AgentConfigDict,
    model_factory_func: ModelFactoryFunctionProtocol,
    group: list[str],
    active_agents: AgentDict,
):
    """
    For each of the possible agent types, create agents and update the active agents dictionary.

    Notice: This function modifies the active_agents dictionary in-place which is a feature (no copies).
    """
    type_of_agents = list(map(lambda c: (c, c.value), AgentType))
    for agent_type, agent_type_value in type_of_agents:
        agents_dict_local = {
            name: agents_dict[name]
            for name in group
            if normalize_agent_type_value(agents_dict[name].get("type")) == agent_type_value
        }
        agents = await create_agents_by_type(
            app_instance,
            agents_dict_local,
            agent_type,
            model_factory_func,
            active_agents,
        )
        active_agents.update(agents)


async def create_agents_in_dependency_order(
    app_instance: CoreContextProtocol,
    agents_dict: AgentConfigDict,
    model_factory_func: ModelFactoryFunctionProtocol,
    allow_cycles: bool = False,
) -> AgentDict:
    """
    Create agent instances in dependency order without proxies.

    Args:
        app_instance: The main application instance
        agents_dict: Dictionary of agent configurations
        model_factory_func: Function for creating model factories
        allow_cycles: Whether to allow cyclic dependencies

    Returns:
        Dictionary of initialized agent instances
    """
    # Get the dependencies between agents
    dependencies = get_dependencies_groups(agents_dict, allow_cycles)

    # Create a dictionary to store all active agents/workflows
    active_agents: AgentDict = {}

    active_agents_in_dependency_group_partial = partial(
        active_agents_in_dependency_group,
        app_instance,
        agents_dict,
        model_factory_func,
    )

    # Create agent proxies for each group in dependency order
    for group in dependencies:
        await active_agents_in_dependency_group_partial(group, active_agents)

    return active_agents


async def create_basic_agents_in_dependency_order(
    context: Context,
    agents_dict: AgentConfigDict,
    model_factory_func: ModelFactoryFunctionProtocol,
    allow_cycles: bool = False,
) -> AgentDict:
    """
    Create BASIC agents in dependency order without a full Core instance.

    Args:
        context: Application context
        agents_dict: Dictionary of agent configurations
        model_factory_func: Function for creating model factories
        allow_cycles: Whether to allow cyclic dependencies

    Returns:
        Dictionary of initialized agent instances
    """
    _ensure_basic_only_agents(agents_dict)
    return await create_agents_in_dependency_order(
        _ContextCoreShim(context),
        agents_dict,
        model_factory_func,
        allow_cycles,
    )


async def _create_default_fan_in_agent(
    fan_in_name: str,
    context,
    model_factory_func: ModelFactoryFunctionProtocol,
) -> AgentProtocol:
    """
    Create a default fan-in agent for parallel workflows when none is specified.

    Args:
        fan_in_name: Name for the new fan-in agent
        context: Application context
        model_factory_func: Function for creating model factories

    Returns:
        Initialized Agent instance for fan-in operations
    """
    # Create a simple config for the fan-in agent with passthrough model
    default_config = AgentConfig(
        name=fan_in_name,
        model="passthrough",
        instruction="You are a passthrough agent that combines outputs from parallel agents.",
    )

    # Create and initialize the default agent
    fan_in_agent = LlmAgent(
        config=default_config,
        context=context,
    )
    await fan_in_agent.initialize()

    # Attach LLM to the agent
    llm_factory = model_factory_func(model="passthrough")
    await fan_in_agent.attach_llm(llm_factory)

    return fan_in_agent



"""Local model overlay discovery and resolution."""

from __future__ import annotations

import os
from dataclasses import dataclass
from pathlib import Path
from typing import Literal
from urllib.parse import urlencode

import yaml
from pydantic import AliasChoices, BaseModel, ConfigDict, Field, model_validator

import fast_agent.config as config_module
from fast_agent.config import load_yaml_mapping, resolve_environment_config_file
from fast_agent.core.exceptions import ModelConfigError
from fast_agent.core.logging.logger import get_logger
from fast_agent.llm.model_database import ModelDatabase, ModelParameters
from fast_agent.llm.provider_types import Provider

logger = get_logger(__name__)


def _normalize_reasoning_value(value: str | bool | int | None) -> str | None:
    if value is None:
        return None
    if isinstance(value, bool):
        return "on" if value else "off"
    if isinstance(value, int):
        return str(value)
    normalized = value.strip()
    return normalized or None


def _normalize_toggle_value(value: bool | None) -> str | None:
    if value is None:
        return None
    return "on" if value else "off"


def _prefixed_model_spec(provider: Provider, model_name: str) -> str:
    if model_name.startswith(f"{provider.value}.") or model_name.startswith(f"{provider.value}/"):
        return model_name
    return f"{provider.value}.{model_name}"


def _overlay_model_key(provider: Provider, model_name: str) -> str:
    normalized = model_name.strip().lower()
    if provider == Provider.HUGGINGFACE and ":" in normalized:
        return normalized.rsplit(":", 1)[0]
    return normalized


def _existing_model_params(provider: Provider, model_name: str) -> ModelParameters | None:
    normalized = _overlay_model_key(provider, model_name)
    params = ModelDatabase.MODELS.get(normalized)
    if params is not None:
        return params
    return ModelDatabase._RUNTIME_MODEL_PARAMS.get(normalized)


class ModelOverlayConnection(BaseModel):
    """Connection overrides attached to a local model overlay."""

    model_config = ConfigDict(extra="ignore")

    base_url: str | None = None
    auth: Literal["none", "env", "secret_ref"] | None = None
    api_key_env: str | None = None
    secret_ref: str | None = None
    default_headers: dict[str, str] = Field(default_factory=dict)

    @model_validator(mode="after")
    def _validate_auth_configuration(self) -> "ModelOverlayConnection":
        if self.auth == "env" and not self.api_key_env:
            raise ValueError("connection.api_key_env is required when connection.auth is 'env'")
        if self.auth == "secret_ref" and not self.secret_ref:
            raise ValueError("connection.secret_ref is required when connection.auth is 'secret_ref'")
        return self

    def auth_mode(self) -> Literal["none", "env", "secret_ref"] | None:
        if self.auth is not None:
            return self.auth
        if self.api_key_env:
            return "env"
        if self.secret_ref:
            return "secret_ref"
        return None


class ModelOverlayDefaults(BaseModel):
    """Request defaults applied by a local overlay."""

    model_config = ConfigDict(extra="ignore")

    reasoning: str | bool | int | None = None
    temperature: float | None = None
    top_p: float | None = None
    top_k: int | None = None
    min_p: float | None = None
    presence_penalty: float | None = None
    repetition_penalty: float | None = None
    transport: Literal["sse", "websocket", "auto"] | None = None
    service_tier: Literal["fast", "flex"] | None = None
    web_search: bool | None = None
    web_fetch: bool | None = None
    max_tokens: int | None = Field(
        default=None,
        validation_alias=AliasChoices("max_tokens", "maxTokens"),
    )

    def to_query_pairs(self) -> list[tuple[str, str]]:
        pairs: list[tuple[str, str]] = []
        reasoning = _normalize_reasoning_value(self.reasoning)
        if reasoning is not None:
            pairs.append(("reasoning", reasoning))
        if self.temperature is not None:
            pairs.append(("temperature", str(self.temperature)))
        if self.top_p is not None:
            pairs.append(("top_p", str(self.top_p)))
        if self.top_k is not None:
            pairs.append(("top_k", str(self.top_k)))
        if self.min_p is not None:
            pairs.append(("min_p", str(self.min_p)))
        if self.presence_penalty is not None:
            pairs.append(("presence_penalty", str(self.presence_penalty)))
        if self.repetition_penalty is not None:
            pairs.append(("repetition_penalty", str(self.repetition_penalty)))
        if self.transport is not None:
            pairs.append(("transport", self.transport))
        if self.service_tier is not None:
            pairs.append(("service_tier", self.service_tier))
        web_search = _normalize_toggle_value(self.web_search)
        if web_search is not None:
            pairs.append(("web_search", web_search))
        web_fetch = _normalize_toggle_value(self.web_fetch)
        if web_fetch is not None:
            pairs.append(("web_fetch", web_fetch))
        return pairs


class ModelOverlayMetadata(BaseModel):
    """Runtime metadata attached to a local overlay."""

    model_config = ConfigDict(extra="ignore")

    context_window: int | None = None
    max_output_tokens: int | None = None
    tokenizes: list[str] | None = None
    # Legacy fallback retained for older overlay files. New overlays should use
    # defaults.temperature instead.
    default_temperature: float | None = None
    fast: bool | None = None


class ModelOverlayPicker(BaseModel):
    """Picker presentation metadata for a local overlay."""

    model_config = ConfigDict(extra="ignore")

    label: str | None = None
    description: str | None = None
    current: bool = True
    featured: bool = False


class ModelOverlaySecretEntry(BaseModel):
    """Secret companion entry for a model overlay."""

    model_config = ConfigDict(extra="ignore")

    api_key: str | None = None
    default_headers: dict[str, str] | None = None


class ModelOverlayManifest(BaseModel):
    """User-authored local model overlay manifest."""

    model_config = ConfigDict(extra="ignore")

    name: str
    provider: Provider
    model: str
    connection: ModelOverlayConnection = Field(default_factory=ModelOverlayConnection)
    defaults: ModelOverlayDefaults = Field(
        default_factory=ModelOverlayDefaults,
        validation_alias=AliasChoices("defaults", "request_defaults"),
    )
    metadata: ModelOverlayMetadata = Field(
        default_factory=ModelOverlayMetadata,
        validation_alias=AliasChoices("metadata", "model_metadata"),
    )
    picker: ModelOverlayPicker = Field(default_factory=ModelOverlayPicker)


@dataclass(frozen=True, slots=True)
class LoadedModelOverlay:
    """Loaded overlay plus file provenance and secret companion data."""

    manifest: ModelOverlayManifest
    manifest_path: Path
    secret_entry: ModelOverlaySecretEntry | None = None

    @property
    def name(self) -> str:
        return self.manifest.name

    @property
    def provider(self) -> Provider:
        return self.manifest.provider

    @property
    def model_name(self) -> str:
        return self.manifest.model

    @property
    def display_label(self) -> str:
        label = self.manifest.picker.label
        return label if label is not None and label.strip() else self.name

    @property
    def description(self) -> str | None:
        description = self.manifest.picker.description
        if description is None:
            return None
        normalized = description.strip()
        return normalized or None

    @property
    def current(self) -> bool:
        return self.manifest.picker.current

    @property
    def featured(self) -> bool:
        return self.manifest.picker.featured

    @property
    def fast(self) -> bool:
        return bool(self.manifest.metadata.fast)

    @property
    def compiled_model_spec(self) -> str:
        model_spec = _prefixed_model_spec(self.provider, self.model_name)
        query = self.manifest.defaults.to_query_pairs()
        if not query:
            return model_spec
        return f"{model_spec}?{urlencode(query)}"

    def resolved_default_headers(self) -> dict[str, str] | None:
        headers = dict(self.manifest.connection.default_headers)
        if self.secret_entry and self.secret_entry.default_headers:
            headers.update(self.secret_entry.default_headers)
        return headers or None

    def resolved_api_key(self) -> str | None:
        auth_mode = self.manifest.connection.auth_mode()
        if auth_mode == "none":
            return ""
        if auth_mode == "env":
            env_name = self.manifest.connection.api_key_env
            if not env_name:
                raise ModelConfigError(
                    f"Overlay '{self.name}' is missing connection.api_key_env",
                    "Set connection.api_key_env or change connection.auth.",
                )
            api_key = os.getenv(env_name)
            if api_key is None:
                raise ModelConfigError(
                    f"Overlay '{self.name}' requires environment variable '{env_name}'",
                    f"Set {env_name} before using overlay '{self.name}'.",
                )
            return api_key
        if auth_mode == "secret_ref":
            secret_ref = self.manifest.connection.secret_ref
            if not secret_ref:
                raise ModelConfigError(
                    f"Overlay '{self.name}' is missing connection.secret_ref",
                    "Set connection.secret_ref or change connection.auth.",
                )
            if self.secret_entry is None or self.secret_entry.api_key is None:
                raise ModelConfigError(
                    f"Overlay '{self.name}' secret ref '{secret_ref}' could not be resolved",
                    "Add an api_key entry to .fast-agent/model-overlays.secrets.yaml.",
                )
            return self.secret_entry.api_key
        return None

    def llm_init_kwargs(self) -> dict[str, object]:
        kwargs: dict[str, object] = {}
        if self.manifest.connection.base_url is not None:
            kwargs["base_url"] = self.manifest.connection.base_url
        api_key = self.resolved_api_key()
        if api_key is not None:
            kwargs["api_key"] = api_key
        default_headers = self.resolved_default_headers()
        if default_headers is not None:
            kwargs["default_headers"] = default_headers
        return kwargs

    def build_model_parameters(self) -> ModelParameters | None:
        existing = _existing_model_params(self.provider, self.model_name)
        context_window = self.manifest.metadata.context_window
        if context_window is None and existing is not None:
            context_window = existing.context_window

        max_output_tokens = self.manifest.metadata.max_output_tokens
        if max_output_tokens is None:
            max_output_tokens = self.manifest.defaults.max_tokens
        if max_output_tokens is None and existing is not None:
            max_output_tokens = existing.max_output_tokens

        if context_window is None or max_output_tokens is None:
            return None

        default_temperature = self.manifest.defaults.temperature
        if default_temperature is None:
            default_temperature = self.manifest.metadata.default_temperature

        if existing is not None:
            update_payload: dict[str, object] = {
                "context_window": context_window,
                "max_output_tokens": max_output_tokens,
                "default_provider": self.provider,
            }
            if self.manifest.metadata.tokenizes is not None:
                update_payload["tokenizes"] = self.manifest.metadata.tokenizes
            if default_temperature is not None:
                update_payload["default_temperature"] = default_temperature
            if self.manifest.metadata.fast is not None:
                update_payload["fast"] = self.manifest.metadata.fast
            return existing.model_copy(update=update_payload)

        tokenizes = self.manifest.metadata.tokenizes or list(ModelDatabase.TEXT_ONLY)
        return ModelParameters(
            context_window=context_window,
            max_output_tokens=max_output_tokens,
            tokenizes=tokenizes,
            default_provider=self.provider,
            default_temperature=default_temperature,
            fast=bool(self.manifest.metadata.fast),
        )


@dataclass(frozen=True, slots=True)
class ModelOverlayRegistry:
    """Resolved local overlay registry."""

    overlays: tuple[LoadedModelOverlay, ...]
    env_root: Path

    def by_name(self) -> dict[str, LoadedModelOverlay]:
        return {overlay.name: overlay for overlay in self.overlays}

    def runtime_presets(self) -> dict[str, str]:
        return {overlay.name: overlay.compiled_model_spec for overlay in self.overlays}

    def providers(self) -> tuple[Provider, ...]:
        seen: set[Provider] = set()
        ordered: list[Provider] = []
        for overlay in self.overlays:
            if overlay.provider in seen:
                continue
            seen.add(overlay.provider)
            ordered.append(overlay.provider)
        return tuple(ordered)

    def entries_for_provider(self, provider: Provider) -> tuple[LoadedModelOverlay, ...]:
        overlays = [overlay for overlay in self.overlays if overlay.provider == provider]
        overlays.sort(key=lambda overlay: (not overlay.current, not overlay.featured, overlay.name))
        return tuple(overlays)

    def resolve_model_string(self, model_string: str) -> LoadedModelOverlay | None:
        raw_token = model_string.partition("?")[0].strip()
        if not raw_token:
            return None
        return self.by_name().get(raw_token)


@dataclass(frozen=True, slots=True)
class ModelOverlayPaths:
    """Filesystem paths used by the model overlay registry."""

    env_root: Path
    overlays_dir: Path
    secrets_path: Path


def _load_secret_entries(path: Path) -> dict[str, ModelOverlaySecretEntry]:
    payload = load_yaml_mapping(path)
    if not payload:
        return {}

    raw_entries = payload.get("overlays")
    if raw_entries is None:
        raw_entries = payload
    if not isinstance(raw_entries, dict):
        return {}

    entries: dict[str, ModelOverlaySecretEntry] = {}
    for name, entry_payload in raw_entries.items():
        if not isinstance(name, str) or not isinstance(entry_payload, dict):
            continue
        try:
            entries[name] = ModelOverlaySecretEntry.model_validate(entry_payload)
        except Exception as exc:
            logger.warning(
                "Skipping invalid model overlay secret entry",
                secret_ref=name,
                path=str(path),
                error=str(exc),
            )
    return entries


def _load_overlay_file(
    path: Path,
    *,
    secret_entries: dict[str, ModelOverlaySecretEntry],
) -> LoadedModelOverlay | None:
    payload = load_yaml_mapping(path)
    if not payload:
        return None

    try:
        manifest = ModelOverlayManifest.model_validate(payload)
    except Exception as exc:
        logger.warning(
            "Skipping invalid model overlay manifest",
            path=str(path),
            error=str(exc),
        )
        return None

    secret_entry = None
    secret_ref = manifest.connection.secret_ref
    if secret_ref:
        secret_entry = secret_entries.get(secret_ref)

    return LoadedModelOverlay(
        manifest=manifest,
        manifest_path=path,
        secret_entry=secret_entry,
    )


def resolve_model_overlay_paths(
    *,
    start_path: Path | None = None,
    env_dir: str | Path | None = None,
) -> ModelOverlayPaths:
    """Resolve the active model overlay storage paths."""

    base_path = (start_path or Path.cwd()).resolve()
    override = env_dir
    if override is None:
        override = os.getenv("ENVIRONMENT_DIR")
    if override is None:
        configured = _settings_environment_override(start_path=start_path)
        if configured is not None:
            base_path, override = configured

    env_root = resolve_environment_config_file(base_path, env_dir=override).parent
    return ModelOverlayPaths(
        env_root=env_root,
        overlays_dir=env_root / "model-overlays",
        secrets_path=env_root / "model-overlays.secrets.yaml",
    )


def load_model_overlay_secret_entries(
    *,
    start_path: Path | None = None,
    env_dir: str | Path | None = None,
) -> dict[str, ModelOverlaySecretEntry]:
    """Load companion overlay secret entries from the active environment."""

    paths = resolve_model_overlay_paths(start_path=start_path, env_dir=env_dir)
    return _load_secret_entries(paths.secrets_path)


def serialize_model_overlay_manifest(manifest: ModelOverlayManifest) -> str:
    """Serialize a model overlay manifest to YAML."""

    payload = manifest.model_dump(mode="json", exclude_none=True)
    return f"{yaml.safe_dump(payload, sort_keys=False).rstrip()}\n"


def write_model_overlay_manifest(
    manifest: ModelOverlayManifest,
    *,
    start_path: Path | None = None,
    env_dir: str | Path | None = None,
    replace: bool = False,
) -> Path:
    """Write a model overlay manifest into the active environment directory."""

    paths = resolve_model_overlay_paths(start_path=start_path, env_dir=env_dir)
    paths.overlays_dir.mkdir(parents=True, exist_ok=True)
    output_path = paths.overlays_dir / f"{_safe_overlay_filename(manifest.name)}.yaml"
    if output_path.exists() and not replace:
        raise FileExistsError(f"Model overlay already exists at {output_path}")

    output_path.write_text(
        serialize_model_overlay_manifest(manifest),
        encoding="utf-8",
    )
    return output_path


def _safe_overlay_filename(name: str) -> str:
    normalized = name.strip()
    if not normalized:
        return "overlay"

    safe_chars: list[str] = []
    last_was_dash = False
    for char in normalized:
        if char.isalnum() or char in {"-", "_", "."}:
            safe_chars.append(char)
            last_was_dash = False
            continue
        if last_was_dash:
            continue
        safe_chars.append("-")
        last_was_dash = True
    safe_name = "".join(safe_chars).strip("-.")
    return safe_name or "overlay"


def _settings_environment_override(
    *,
    start_path: Path | None,
) -> tuple[Path, str | Path | None] | None:
    settings = getattr(config_module, "_settings", None)
    if settings is None:
        return None

    environment_dir = getattr(settings, "environment_dir", None)
    if environment_dir is None:
        return None

    base_path = (start_path or Path.cwd()).resolve()
    config_file = getattr(settings, "_config_file", None)
    if start_path is None and isinstance(config_file, str) and config_file.strip():
        base_path = Path(config_file).expanduser().resolve().parent

    return base_path, environment_dir


def load_model_overlay_registry(
    *,
    start_path: Path | None = None,
    env_dir: str | Path | None = None,
) -> ModelOverlayRegistry:
    """Load static model overlays from the active environment directory."""

    paths = resolve_model_overlay_paths(start_path=start_path, env_dir=env_dir)
    secret_entries = _load_secret_entries(paths.secrets_path)

    loaded: dict[str, LoadedModelOverlay] = {}
    if paths.overlays_dir.exists():
        for path in sorted(paths.overlays_dir.iterdir()):
            if not path.is_file() or path.suffix.lower() not in {".yaml", ".yml"}:
                continue
            overlay = _load_overlay_file(path, secret_entries=secret_entries)
            if overlay is None:
                continue
            if overlay.name in loaded:
                logger.warning(
                    "Duplicate model overlay name detected; replacing earlier manifest",
                    overlay_name=overlay.name,
                    replaced_path=str(loaded[overlay.name].manifest_path),
                    path=str(path),
                )
            loaded[overlay.name] = overlay

    return ModelOverlayRegistry(
        overlays=tuple(loaded.values()),
        env_root=paths.env_root,
    )



"""Pure command-line parsing for interactive prompt input."""

from __future__ import annotations

import os
import platform
import shlex
import shutil
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable

from fast_agent.commands.mcp_command_intents import (
    parse_mcp_connect_tokens,
    parse_mcp_session_tokens,
)
from fast_agent.commands.shared_command_intents import (
    parse_current_agent_history_intent,
    parse_session_command_intent,
)
from fast_agent.ui.command_payloads import (
    AgentCommand,
    CardsCommand,
    ClearCommand,
    ClearSessionsCommand,
    CommandPayload,
    CreateSessionCommand,
    ForkSessionCommand,
    HashAgentCommand,
    HistoryFixCommand,
    HistoryReviewCommand,
    HistoryRewindCommand,
    HistoryShowCommand,
    HistoryWebClearCommand,
    ListSessionsCommand,
    ListToolsCommand,
    LoadAgentCardCommand,
    LoadHistoryCommand,
    LoadPromptCommand,
    McpConnectCommand,
    McpConnectMode,
    McpDisconnectCommand,
    McpListCommand,
    McpReconnectCommand,
    ModelFastCommand,
    ModelReasoningCommand,
    ModelsCommand,
    ModelSwitchCommand,
    ModelVerbosityCommand,
    ModelWebFetchCommand,
    ModelWebSearchCommand,
    PinSessionCommand,
    ReloadAgentsCommand,
    ResumeSessionCommand,
    SaveHistoryCommand,
    SelectPromptCommand,
    ShellCommand,
    ShowHistoryCommand,
    ShowMarkdownCommand,
    ShowMcpStatusCommand,
    ShowSystemCommand,
    ShowUsageCommand,
    SkillsCommand,
    SwitchAgentCommand,
    TitleSessionCommand,
    UnknownCommand,
)


def _default_shell_command() -> str:
    if platform.system() == "Windows":
        for shell_name in ["pwsh", "powershell", "cmd"]:
            shell_path = shutil.which(shell_name)
            if shell_path:
                return shell_path
        return os.environ.get("COMSPEC", "cmd.exe")

    shell_env = os.environ.get("SHELL")
    if shell_env and Path(shell_env).exists():
        return shell_env

    for shell_name in ["bash", "zsh", "sh"]:
        shell_path = shutil.which(shell_name)
        if shell_path:
            return shell_path

    return "sh"


def _infer_mcp_connect_mode(target_text: str) -> McpConnectMode:
    stripped = target_text.strip().lower()
    if stripped.startswith(("http://", "https://")):
        return "url"
    if stripped.startswith("@"):
        return "npx"
    if stripped.startswith("npx "):
        return "npx"
    if stripped.startswith("uvx "):
        return "uvx"
    return "stdio"


def _rebuild_mcp_target_text(tokens: list[str]) -> str:
    if not tokens:
        return ""

    rebuilt_parts: list[str] = []
    for token in tokens:
        if token == "" or any(char.isspace() for char in token):
            rebuilt_parts.append(shlex.quote(token))
        else:
            rebuilt_parts.append(token)
    return " ".join(rebuilt_parts)


def _parse_quoted_history_target(text: str) -> str | None:
    stripped = text.strip()
    if not stripped:
        return None

    try:
        tokens = shlex.split(stripped)
    except ValueError:
        return None

    if len(tokens) != 1:
        return None

    # Allow explicit quoting/escaping to force agent-name parsing for values
    # that would otherwise collide with /history subcommands.
    if stripped == tokens[0]:
        return None
    return tokens[0]


def _parse_mcp_single_server_name(tokens: list[str], *, usage: str) -> tuple[str | None, str | None]:
    name = tokens[1] if len(tokens) > 1 else None
    error = None if name else usage
    return name, error


def _parse_hash_agent_command(body: str, *, quiet: bool) -> HashAgentCommand | str:
    stripped = body.strip()
    if not stripped:
        prefix = "##" if quiet else "#"
        return f"{prefix}{body}"

    for index, char in enumerate(stripped):
        if char.isspace():
            agent_name = stripped[:index]
            message = stripped[index:].strip()
            return HashAgentCommand(agent_name=agent_name, message=message, quiet=quiet)

    return HashAgentCommand(agent_name=stripped, message="", quiet=quiet)


def _parse_history_command(remainder: str) -> CommandPayload:
    if not remainder:
        return ShowHistoryCommand(agent=None)

    quoted_target = _parse_quoted_history_target(remainder)
    if quoted_target is not None:
        return ShowHistoryCommand(agent=quoted_target)

    try:
        tokens = shlex.split(remainder)
    except ValueError:
        candidate = remainder.strip()
        return ShowHistoryCommand(agent=candidate or None)

    if not tokens:
        return ShowHistoryCommand(agent=None)

    subcmd = tokens[0].lower()
    argument = " ".join(tokens[1:]).strip()

    targeted_payload = _parse_targeted_history_action(subcmd, argument)
    if targeted_payload is not None:
        return targeted_payload

    intent = parse_current_agent_history_intent(remainder)
    shared_payload = _history_payload_from_shared_intent(intent)
    if shared_payload is not None:
        return shared_payload

    return ShowHistoryCommand(agent=remainder)


def _parse_targeted_history_action(subcmd: str, argument: str) -> CommandPayload | None:
    if subcmd == "rewind":
        return _parse_history_rewind_command(argument)
    if subcmd == "fix":
        return HistoryFixCommand(agent=argument or None)
    if subcmd == "webclear":
        return HistoryWebClearCommand(agent=argument or None)
    if subcmd == "clear":
        return _parse_history_clear_command(argument)
    return None


def _parse_history_rewind_command(argument: str) -> HistoryRewindCommand:
    stripped = argument.strip()
    if not stripped:
        return HistoryRewindCommand(
            turn_index=None,
            error="Turn number required for /history rewind",
        )
    try:
        turn_index = int(stripped)
    except ValueError:
        return HistoryRewindCommand(
            turn_index=None,
            error="Turn number must be an integer",
        )
    return HistoryRewindCommand(turn_index=turn_index, error=None)


def _history_payload_from_shared_intent(intent) -> CommandPayload | None:
    if intent.action == "overview":
        return ShowHistoryCommand(agent=None)
    if intent.action == "show":
        return HistoryShowCommand(agent=intent.argument)
    if intent.action == "save":
        return SaveHistoryCommand(filename=intent.argument)
    if intent.action == "load":
        if not intent.argument:
            return LoadHistoryCommand(
                filename=None,
                error="Filename required for /history load",
            )
        return LoadHistoryCommand(filename=intent.argument, error=None)
    if intent.action == "detail":
        return _history_review_payload_from_intent(intent.turn_index, intent.turn_error)
    return None


def _history_review_payload_from_intent(
    turn_index: int | None,
    turn_error: str | None,
) -> HistoryReviewCommand:
    if turn_error == "missing":
        return HistoryReviewCommand(
            turn_index=None,
            error="Turn number required for /history detail",
        )
    if turn_error == "invalid":
        return HistoryReviewCommand(
            turn_index=None,
            error="Turn number must be an integer",
        )
    return HistoryReviewCommand(turn_index=turn_index, error=None)


def _parse_history_clear_command(argument: str) -> ClearCommand:
    clear_tokens = argument.split(maxsplit=1) if argument else []
    action = clear_tokens[0].lower() if clear_tokens else "all"
    target_agent = clear_tokens[1].strip() if len(clear_tokens) > 1 else None
    if action == "last":
        return ClearCommand(kind="clear_last", agent=target_agent)
    if action == "all":
        return ClearCommand(kind="clear_history", agent=target_agent)
    return ClearCommand(kind="clear_history", agent=argument or None)


def _parse_session_command(remainder: str) -> CommandPayload:
    intent = parse_session_command_intent(remainder)
    if intent.action in {"help", "unknown"}:
        return ListSessionsCommand(show_help=True)
    if intent.action == "list":
        return ListSessionsCommand()
    if intent.action == "new":
        return CreateSessionCommand(session_name=intent.argument)
    if intent.action == "resume":
        return ResumeSessionCommand(session_id=intent.argument)
    if intent.action == "title":
        return TitleSessionCommand(title=intent.argument or "")
    if intent.action == "fork":
        return ForkSessionCommand(title=intent.argument)
    if intent.action == "delete":
        return ClearSessionsCommand(target=intent.argument)
    return PinSessionCommand(value=intent.pin_value, target=intent.pin_target)


def _parse_card_command(remainder: str) -> CommandPayload:
    if not remainder:
        return LoadAgentCardCommand(
            filename=None,
            add_tool=False,
            remove_tool=False,
            error="Filename required for /card",
        )

    try:
        tokens = shlex.split(remainder)
    except ValueError as exc:
        return LoadAgentCardCommand(
            filename=None,
            add_tool=False,
            remove_tool=False,
            error=f"Invalid arguments: {exc}",
        )

    add_tool = False
    remove_tool = False
    filename = None
    for token in tokens:
        if token in {"tool", "--tool", "--as-tool", "-t"}:
            add_tool = True
            continue
        if token in {"remove", "--remove", "--rm"}:
            remove_tool = True
            add_tool = True
            continue
        if filename is None:
            filename = token

    if not filename:
        return LoadAgentCardCommand(
            filename=None,
            add_tool=add_tool,
            remove_tool=remove_tool,
            error="Filename required for /card",
        )

    return LoadAgentCardCommand(
        filename=filename,
        add_tool=add_tool,
        remove_tool=remove_tool,
        error=None,
    )


def _parse_agent_command(remainder: str) -> CommandPayload:
    if not remainder:
        return AgentCommand(
            agent_name=None,
            add_tool=False,
            remove_tool=False,
            dump=False,
            error="Usage: /agent <name> --tool | /agent [name] --dump",
        )

    try:
        tokens = shlex.split(remainder)
    except ValueError as exc:
        return AgentCommand(
            agent_name=None,
            add_tool=False,
            remove_tool=False,
            dump=False,
            error=f"Invalid arguments: {exc}",
        )

    agent_command = _parse_agent_tokens(tokens)
    return _validate_agent_command(agent_command)


def _parse_agent_tokens(tokens: list[str]) -> AgentCommand:
    add_tool = False
    remove_tool = False
    dump = False
    agent_name = None
    unknown: list[str] = []

    for token in tokens:
        if token in {"tool", "--tool", "--as-tool", "-t"}:
            add_tool = True
            continue
        if token in {"remove", "--remove", "--rm"}:
            remove_tool = True
            add_tool = True
            continue
        if token in {"dump", "--dump", "-d"}:
            dump = True
            continue
        if agent_name is None:
            agent_name = token[1:] if token.startswith("@") else token
            continue
        unknown.append(token)

    error = f"Unexpected arguments: {', '.join(unknown)}" if unknown else None
    return AgentCommand(
        agent_name=agent_name,
        add_tool=add_tool,
        remove_tool=remove_tool,
        dump=dump,
        error=error,
    )


def _validate_agent_command(command: AgentCommand) -> AgentCommand:
    if command.error is not None:
        return command
    if command.add_tool and command.dump:
        return AgentCommand(
            agent_name=command.agent_name,
            add_tool=command.add_tool,
            remove_tool=command.remove_tool,
            dump=command.dump,
            error="Use either --tool or --dump, not both",
        )
    if not command.add_tool and not command.dump:
        return AgentCommand(
            agent_name=command.agent_name,
            add_tool=command.add_tool,
            remove_tool=command.remove_tool,
            dump=command.dump,
            error="Usage: /agent <name> --tool | /agent [name] --dump",
        )
    if command.add_tool and not command.agent_name:
        return AgentCommand(
            agent_name=command.agent_name,
            add_tool=command.add_tool,
            remove_tool=command.remove_tool,
            dump=command.dump,
            error="Agent name is required for /agent --tool",
        )
    return command


def _parse_mcp_command(remainder: str) -> CommandPayload:
    if not remainder:
        return ShowMcpStatusCommand()

    try:
        tokens = shlex.split(remainder)
    except ValueError as exc:
        return McpConnectCommand(
            target_text="",
            parsed_mode="stdio",
            server_name=None,
            auth_token=None,
            timeout_seconds=None,
            trigger_oauth=None,
            reconnect_on_disconnect=None,
            force_reconnect=False,
            error=f"Invalid arguments: {exc}",
        )

    subcmd = tokens[0].lower() if tokens else ""
    if subcmd == "list":
        return McpListCommand()
    if subcmd == "disconnect":
        name, error = _parse_mcp_single_server_name(
            tokens,
            usage="Usage: /mcp disconnect <server_name>",
        )
        return McpDisconnectCommand(server_name=name, error=error)
    if subcmd == "reconnect":
        name, error = _parse_mcp_single_server_name(
            tokens,
            usage="Usage: /mcp reconnect <server_name>",
        )
        return McpReconnectCommand(server_name=name, error=error)
    if subcmd == "session":
        return parse_mcp_session_tokens(tokens[1:])
    if subcmd == "connect":
        return parse_mcp_connect_tokens(tokens[1:])
    return UnknownCommand(command="mcp")


def _parse_connect_alias_command(remainder: str) -> McpConnectCommand:
    parsed_mode = _infer_mcp_connect_mode(remainder)
    if not remainder:
        return McpConnectCommand(
            target_text="",
            parsed_mode="stdio",
            server_name=None,
            auth_token=None,
            timeout_seconds=None,
            trigger_oauth=None,
            reconnect_on_disconnect=None,
            force_reconnect=False,
            error="Usage: /connect <target>",
        )
    return McpConnectCommand(
        target_text=remainder,
        parsed_mode=parsed_mode,
        server_name=None,
        auth_token=None,
        timeout_seconds=None,
        trigger_oauth=None,
        reconnect_on_disconnect=None,
        force_reconnect=False,
        error=None,
    )


def _parse_prompt_command(remainder: str) -> CommandPayload:
    if not remainder:
        return SelectPromptCommand(prompt_index=None, prompt_name=None)

    try:
        tokens = shlex.split(remainder)
    except ValueError:
        tokens = []

    if tokens:
        subcmd = tokens[0].lower()
        argument = remainder[len(tokens[0]) :].strip()
        if subcmd == "load":
            if not argument:
                return LoadPromptCommand(filename=None, error="Filename required for /prompt load")
            return LoadPromptCommand(filename=argument, error=None)

    if remainder.lower().endswith((".json", ".md")):
        return LoadPromptCommand(filename=remainder, error=None)
    if remainder.isdigit():
        return SelectPromptCommand(prompt_index=int(remainder), prompt_name=None)
    return SelectPromptCommand(prompt_index=None, prompt_name=remainder)


def _parse_model_command(cmd_line: str, remainder: str) -> CommandPayload:
    if not remainder:
        return ModelReasoningCommand(value=None)

    try:
        tokens = shlex.split(remainder)
    except ValueError:
        tokens = remainder.split(maxsplit=1)

    if not tokens:
        return ModelReasoningCommand(value=None)

    subcmd = tokens[0].lower()
    argument = remainder[len(tokens[0]) :].strip()

    value = argument or None
    value_command_factories: dict[str, Callable[[str | None], CommandPayload]] = {
        "reasoning": ModelReasoningCommand,
        "verbosity": ModelVerbosityCommand,
        "fast": ModelFastCommand,
        "web_search": ModelWebSearchCommand,
        "web_fetch": ModelWebFetchCommand,
        "switch": ModelSwitchCommand,
    }
    factory = value_command_factories.get(subcmd)
    if factory is not None:
        return factory(value)
    if subcmd in {"doctor", "references", "catalog", "help"}:
        return ModelsCommand(action=subcmd, argument=value)
    return UnknownCommand(command=cmd_line)


def _parse_slash_alias_command(
    cmd: str,
    remainder: str,
    *,
    cmd_line: str,
) -> str | CommandPayload | None:
    if cmd in {"save_history", "save"}:
        filename = remainder or None
        return SaveHistoryCommand(filename=filename)
    if cmd in {"load_history", "load"}:
        if not remainder:
            return LoadHistoryCommand(filename=None, error="Filename required for /history load")
        return LoadHistoryCommand(filename=remainder, error=None)
    if cmd == "resume":
        return ResumeSessionCommand(session_id=remainder or None)
    if cmd == "fast":
        return ModelFastCommand(value=remainder or None)
    if cmd == "skills":
        if not remainder:
            return SkillsCommand(action="list", argument=None)
        tokens = remainder.split(maxsplit=1)
        action = tokens[0].lower()
        argument = tokens[1].strip() if len(tokens) > 1 else None
        return SkillsCommand(action=action, argument=argument)
    if cmd == "cards":
        if not remainder:
            return CardsCommand(action="list", argument=None)
        tokens = remainder.split(maxsplit=1)
        action = tokens[0].lower()
        argument = tokens[1].strip() if len(tokens) > 1 else None
        return CardsCommand(action=action, argument=argument)
    return None


def _parse_slash_command(cmd_line: str) -> str | CommandPayload:
    cmd_parts = cmd_line[1:].strip().split(maxsplit=1)
    cmd = cmd_parts[0].lower()
    remainder = cmd_parts[1].strip() if len(cmd_parts) > 1 else ""

    simple_factories: dict[str, Callable[[], str | CommandPayload]] = {
        "help": lambda: "HELP",
        "system": ShowSystemCommand,
        "usage": ShowUsageCommand,
        "markdown": ShowMarkdownCommand,
        "reload": ReloadAgentsCommand,
        "mcpstatus": ShowMcpStatusCommand,
        "tools": ListToolsCommand,
        "exit": lambda: "EXIT",
        "stop": lambda: "STOP",
    }
    simple_factory = simple_factories.get(cmd)
    if simple_factory is not None:
        return simple_factory()

    command_parsers: dict[str, Callable[[str], CommandPayload]] = {
        "history": _parse_history_command,
        "session": _parse_session_command,
        "card": _parse_card_command,
        "agent": _parse_agent_command,
        "mcp": _parse_mcp_command,
        "connect": _parse_connect_alias_command,
        "prompt": _parse_prompt_command,
    }
    parser = command_parsers.get(cmd)
    if parser is not None:
        return parser(remainder)
    if cmd == "model":
        return _parse_model_command(cmd_line, remainder)

    alias_result = _parse_slash_alias_command(cmd, remainder, cmd_line=cmd_line)
    if alias_result is not None:
        return alias_result

    return UnknownCommand(command=cmd_line)


def parse_special_input(text: str) -> str | CommandPayload:
    stripped = text.lstrip()
    cmd_line = stripped.splitlines()[0] if stripped.startswith("/") else text

    if cmd_line and cmd_line.startswith("/"):
        if cmd_line == "/":
            return ""
        return _parse_slash_command(cmd_line)

    if cmd_line and cmd_line.startswith("@"):
        return SwitchAgentCommand(agent_name=cmd_line[1:].strip())

    if cmd_line and cmd_line.startswith("##"):
        quiet_body = cmd_line[2:]
        if quiet_body and not quiet_body[0].isspace():
            return _parse_hash_agent_command(quiet_body, quiet=True)
        return text

    if cmd_line and cmd_line.startswith("#"):
        return _parse_hash_agent_command(cmd_line[1:], quiet=False)

    if cmd_line and cmd_line.startswith("!"):
        command = cmd_line[1:].strip()
        if command:
            return ShellCommand(command=command)
        return ShellCommand(command=_default_shell_command())

    return text



"""Runtime capability helpers for model command handlers."""

from __future__ import annotations


def resolve_web_search_enabled(llm: object) -> bool:
    web_tools_enabled = getattr(llm, "web_tools_enabled", None)
    if isinstance(web_tools_enabled, tuple) and len(web_tools_enabled) >= 1:
        return bool(web_tools_enabled[0])

    enabled = getattr(llm, "web_search_enabled", None)
    if isinstance(enabled, bool):
        return enabled
    return False


def resolve_web_fetch_enabled(llm: object) -> bool:
    web_tools_enabled = getattr(llm, "web_tools_enabled", None)
    if isinstance(web_tools_enabled, tuple) and len(web_tools_enabled) >= 2:
        return bool(web_tools_enabled[1])

    enabled = getattr(llm, "web_fetch_enabled", None)
    if isinstance(enabled, bool):
        return enabled
    return False


def resolve_web_search_supported(llm: object) -> bool:
    supported = getattr(llm, "web_search_supported", None)
    return bool(supported) if isinstance(supported, bool) else False


def resolve_web_fetch_supported(llm: object) -> bool:
    supported = getattr(llm, "web_fetch_supported", None)
    return bool(supported) if isinstance(supported, bool) else False


def set_web_search_enabled(llm: object, value: bool | None) -> None:
    setter = getattr(llm, "set_web_search_enabled", None)
    if callable(setter):
        setter(value)
        return
    raise ValueError("Current model does not support web search configuration.")


def set_web_fetch_enabled(llm: object, value: bool | None) -> None:
    setter = getattr(llm, "set_web_fetch_enabled", None)
    if callable(setter):
        setter(value)
        return
    raise ValueError("Current model does not support web fetch configuration.")


def resolve_service_tier_supported(llm: object) -> bool:
    supported = getattr(llm, "service_tier_supported", None)
    return bool(supported) if isinstance(supported, bool) else False


def available_service_tier_values(llm: object) -> tuple[str, ...]:
    raw_values = getattr(llm, "available_service_tiers", None)
    if isinstance(raw_values, tuple | list):
        values = tuple(value for value in raw_values if value in {"fast", "flex"})
        if values:
            return values
    if resolve_service_tier_supported(llm):
        return ("fast", "flex")
    return ()


def service_tier_command_values(llm: object) -> tuple[str, ...]:
    values = ["on", "off"]
    if "flex" in available_service_tier_values(llm):
        values.append("flex")
    values.append("status")
    return tuple(values)


def resolve_service_tier(llm: object) -> str | None:
    value = getattr(llm, "service_tier", None)
    return value if value in {"fast", "flex"} else None


def set_service_tier(llm: object, value: str | None) -> None:
    setter = getattr(llm, "set_service_tier", None)
    if callable(setter):
        setter(value)
        return
    raise ValueError("Current model does not support service tier configuration.")


def describe_service_tier_state(llm: object) -> str:
    current_tier = resolve_service_tier(llm)
    if current_tier == "fast":
        return "fast"
    if current_tier == "flex":
        return "flex"
    return "default"


def model_supports_web_search(llm: object) -> bool:
    """Return True when model/provider supports web_search runtime configuration."""
    return resolve_web_search_supported(llm)


def model_supports_web_fetch(llm: object) -> bool:
    """Return True when model/provider supports web_fetch runtime configuration."""
    return resolve_web_fetch_supported(llm)


def model_supports_service_tier(llm: object) -> bool:
    """Return True when model/provider supports service tier runtime configuration."""
    return resolve_service_tier_supported(llm)


def model_supports_text_verbosity(llm: object) -> bool:
    """Return True when model exposes text verbosity controls."""
    return getattr(llm, "text_verbosity_spec", None) is not None



"""Prompt key binding helpers."""

from __future__ import annotations

import asyncio
import time
from typing import TYPE_CHECKING, Any

from prompt_toolkit.filters import Condition
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.lexers import Lexer
from rich import print as rich_print

from fast_agent.ui.prompt.editor import get_text_from_editor

if TYPE_CHECKING:
    from collections.abc import Callable

    from prompt_toolkit.buffer import Buffer

    from fast_agent.core.agent_app import AgentApp


class ShellPrefixLexer(Lexer):
    """Lexer that highlights shell (!) and comment (#) commands."""

    def lex_document(self, document):
        def get_line_tokens(line_number):
            line = document.lines[line_number]
            stripped = line.lstrip()
            if stripped.startswith("!"):
                return [("class:shell-command", line)]
            if stripped.startswith("#"):
                return [("class:comment-command", line)]
            return [("", line)]

        return get_line_tokens


class AgentKeyBindings(KeyBindings):
    agent_provider: "AgentApp | None" = None
    current_agent_name: str | None = None


def _cycle_completion(buffer: Buffer, *, backwards: bool) -> bool:
    """Cycle through current completion menu items.

    Returns ``True`` when a completion menu was active and the selection moved.
    """
    state = buffer.complete_state
    if state is None:
        return False

    completions = state.completions
    if not completions:
        return False

    total = len(completions)
    current_index = state.complete_index

    if backwards:
        next_index = total - 1 if current_index is None else (current_index - 1) % total
    else:
        next_index = 0 if current_index is None else (current_index + 1) % total

    buffer.go_to_completion(next_index)
    return True


def _accept_completion(buffer: Buffer) -> bool:
    """Accept the currently highlighted completion without submitting input."""
    state = buffer.complete_state
    if state is None:
        return False

    if not state.completions:
        return False

    if state.complete_index is None:
        buffer.go_to_completion(0)

    # Keep the inserted completion text and close the menu.
    buffer.complete_state = None
    return True


def create_keybindings(
    on_toggle_multiline: Callable[[bool], None] | None = None,
    on_cycle_service_tier: Callable[[], None] | None = None,
    on_cycle_reasoning: Callable[[], None] | None = None,
    on_cycle_verbosity: Callable[[], None] | None = None,
    on_cycle_web_search: Callable[[], None] | None = None,
    on_cycle_web_fetch: Callable[[], None] | None = None,
    app: Any | None = None,
    agent_provider: "AgentApp | None" = None,
    agent_name: str | None = None,
) -> AgentKeyBindings:
    """Create custom key bindings."""
    kb = AgentKeyBindings()

    def _session_state_module():
        from fast_agent.ui.prompt import input as input_state

        return input_state

    def _should_start_completion(text: str) -> bool:
        stripped = text.lstrip()
        if not stripped:
            return True
        if stripped.startswith("!"):
            return True
        if stripped.startswith(("/", "@", "#")):
            return True
        return True

    @kb.add("c-space")
    @kb.add("c-@")
    def _(event) -> None:
        text = event.current_buffer.document.text_before_cursor
        if not _should_start_completion(text):
            return
        event.current_buffer.start_completion()

    @kb.add("tab")
    @kb.add("c-i")
    def _(event) -> None:
        if _cycle_completion(event.current_buffer, backwards=False):
            return

        text = event.current_buffer.document.text_before_cursor
        if not _should_start_completion(text):
            return
        event.current_buffer.start_completion(insert_common_part=True)

    @kb.add("s-tab")
    def _(event) -> None:
        if _cycle_completion(event.current_buffer, backwards=True):
            return

        if on_cycle_service_tier is not None:
            on_cycle_service_tier()
            if event.app:
                event.app.invalidate()
            elif app:
                app.invalidate()
            return

        text = event.current_buffer.document.text_before_cursor
        if not _should_start_completion(text):
            return
        event.current_buffer.start_completion(select_last=True)

    def _invoke_callback(callback: Callable[[], None] | None, event) -> bool:
        if callback is None:
            return False
        callback()
        if event.app:
            event.app.invalidate()
        elif app:
            app.invalidate()
        return True

    @kb.add("f6")
    def _(event) -> None:
        if _invoke_callback(on_cycle_reasoning, event):
            return

    @kb.add("f7")
    def _(event) -> None:
        if _invoke_callback(on_cycle_verbosity, event):
            return

    @kb.add("f8")
    def _(event) -> None:
        if _invoke_callback(on_cycle_web_search, event):
            return

    @kb.add("f9")
    def _(event) -> None:
        if _invoke_callback(on_cycle_web_fetch, event):
            return

    @kb.add("c-m", filter=Condition(lambda: _has_any_completions()), eager=True)
    @kb.add("enter", filter=Condition(lambda: _has_any_completions()), eager=True)
    def _(event) -> None:
        _accept_completion(event.current_buffer)

    def _has_any_completions() -> bool:
        from prompt_toolkit.application.current import get_app

        state = get_app().current_buffer.complete_state
        if state is None:
            return False
        return len(state.completions) > 0

    @kb.add("c-m", filter=Condition(lambda: not _session_state_module().in_multiline_mode))
    def _(event) -> None:
        """Enter: accept input when not in multiline mode."""
        event.current_buffer.validate_and_handle()

    @kb.add("c-j", filter=Condition(lambda: not _session_state_module().in_multiline_mode))
    def _(event) -> None:
        """Ctrl+J: Insert newline when in normal mode."""
        event.current_buffer.insert_text("\n")

    @kb.add("c-m", filter=Condition(lambda: _session_state_module().in_multiline_mode))
    def _(event) -> None:
        """Enter: Insert newline when in multiline mode."""
        event.current_buffer.insert_text("\n")

    @kb.add("c-j", filter=Condition(lambda: _session_state_module().in_multiline_mode))
    def _(event) -> None:
        """Ctrl+J (equivalent to Ctrl+Enter): Submit in multiline mode."""
        event.current_buffer.validate_and_handle()

    @kb.add("c-t")
    def _(event) -> None:
        """Ctrl+T: Toggle multiline mode."""
        session_state = _session_state_module()
        session_state.in_multiline_mode = not session_state.in_multiline_mode

        if event.app:
            event.app.invalidate()
        elif app:
            app.invalidate()

        if on_toggle_multiline:
            on_toggle_multiline(session_state.in_multiline_mode)

    @kb.add("c-l")
    def _(event) -> None:
        """Ctrl+L: Clear and redraw the terminal screen."""
        app_ref = event.app or app
        if app_ref and getattr(app_ref, "renderer", None):
            app_ref.renderer.clear()
            app_ref.invalidate()

    @kb.add("c-u")
    def _(event) -> None:
        """Ctrl+U: Clear the input buffer."""
        event.current_buffer.text = ""

    @kb.add("c-c")
    def _(event) -> None:
        """Ctrl+C: interrupt prompt input (handled by caller policy)."""
        event.app.exit(exception=KeyboardInterrupt())

    @kb.add("c-d")
    def _(event) -> None:
        """Ctrl+D: signal EOF (mapped to STOP by prompt input handler)."""
        event.app.exit(exception=EOFError())

    @kb.add("c-e")
    async def _(event) -> None:
        """Ctrl+E: Edit current buffer in $EDITOR."""
        current_text = event.app.current_buffer.text
        try:
            edited_text = await event.app.loop.run_in_executor(
                None, get_text_from_editor, current_text
            )
            event.app.current_buffer.text = edited_text
            event.app.current_buffer.cursor_position = len(edited_text)
        except asyncio.CancelledError:
            rich_print("[yellow]Editor interaction cancelled.[/yellow]")
        except Exception as exc:
            rich_print(f"[red]Error during editor interaction: {exc}[/red]")
        finally:
            if event.app:
                event.app.invalidate()

    kb.agent_provider = agent_provider
    kb.current_agent_name = agent_name

    @kb.add("c-y")
    async def _(event) -> None:
        """Ctrl+Y: Copy last output (shell or assistant) to clipboard."""

        def _show_copy_notice() -> None:
            session_state = _session_state_module()
            session_state._copy_notice = "COPIED"
            session_state._copy_notice_until = time.monotonic() + 2.0
            if event.app:
                event.app.invalidate()
            else:
                rich_print("\n[green]✓ Copied to clipboard[/green]")

        session_state = _session_state_module()
        if session_state._last_copyable_output:
            try:
                import pyperclip

                pyperclip.copy(session_state._last_copyable_output)
                _show_copy_notice()
                return
            except Exception:
                pass

        if kb.agent_provider and kb.current_agent_name:
            try:
                agent = kb.agent_provider._agent(kb.current_agent_name)
                for msg in reversed(agent.message_history):
                    if msg.role == "assistant":
                        content = msg.last_text()
                        import pyperclip

                        pyperclip.copy(content)
                        _show_copy_notice()
                        return
            except Exception:
                pass

    return kb



"""Skills provenance and sidecar metadata helpers."""

from __future__ import annotations

import hashlib
import json
from datetime import UTC, datetime
from typing import TYPE_CHECKING, Any

from fast_agent.marketplace import formatting as marketplace_formatting
from fast_agent.marketplace import source_utils as marketplace_source_utils
from fast_agent.skills.marketplace_parsing import normalize_repo_path
from fast_agent.skills.models import (
    LOCAL_REVISION,
    SKILL_SOURCE_FILENAME,
    SKILL_SOURCE_SCHEMA_VERSION,
    InstalledSkillSource,
    MarketplaceSkill,
    SkillProvenance,
    SkillSourceOrigin,
)

if TYPE_CHECKING:
    from pathlib import Path


def get_skill_source_sidecar_path(skill_dir: Path) -> Path:
    return skill_dir / SKILL_SOURCE_FILENAME


def compute_skill_content_fingerprint(skill_dir: Path) -> str:
    digest = hashlib.sha256()
    root = skill_dir.resolve()
    sidecar_path = get_skill_source_sidecar_path(root)

    for path in sorted(root.rglob("*")):
        if path == sidecar_path:
            continue
        if not path.is_file():
            continue
        relative = path.relative_to(root).as_posix()
        digest.update(relative.encode("utf-8"))
        digest.update(b"\0")
        digest.update(path.read_bytes())
        digest.update(b"\0")

    return f"sha256:{digest.hexdigest()}"


def read_installed_skill_source(skill_dir: Path) -> tuple[InstalledSkillSource | None, str | None]:
    sidecar_path = get_skill_source_sidecar_path(skill_dir)
    if not sidecar_path.exists():
        return None, None
    try:
        payload = json.loads(sidecar_path.read_text(encoding="utf-8"))
    except Exception as exc:  # noqa: BLE001
        return None, f"invalid json: {exc}"

    if not isinstance(payload, dict):
        return None, "metadata root must be an object"

    try:
        source = parse_installed_skill_source_payload(payload)
    except ValueError as exc:
        return None, str(exc)
    return source, None


def write_installed_skill_source(skill_dir: Path, source: InstalledSkillSource) -> None:
    sidecar_path = get_skill_source_sidecar_path(skill_dir)
    payload = {
        "schema_version": source.schema_version,
        "installed_via": source.installed_via,
        "source_origin": source.source_origin,
        "repo_url": source.repo_url,
        "repo_ref": source.repo_ref,
        "repo_path": source.repo_path,
        "source_url": source.source_url,
        "installed_commit": source.installed_commit,
        "installed_path_oid": source.installed_path_oid,
        "installed_revision": source.installed_revision,
        "installed_at": source.installed_at,
        "content_fingerprint": source.content_fingerprint,
    }
    sidecar_path.write_text(
        json.dumps(payload, indent=2, ensure_ascii=False, sort_keys=True) + "\n",
        encoding="utf-8",
    )


def get_skill_provenance(skill_dir: Path) -> SkillProvenance:
    source, error = read_installed_skill_source(skill_dir)
    if source is None:
        if error is None:
            return SkillProvenance(
                status="unmanaged",
                summary="unmanaged (no sidecar)",
            )
        return SkillProvenance(
            status="invalid_metadata",
            summary=f"invalid metadata ({error})",
            error=error,
        )

    ref_label = f"@{source.repo_ref}" if source.repo_ref else ""
    if source.source_origin == "remote":
        summary = (
            "managed (marketplace)"
            f" • {source.repo_url}{ref_label}"
            f" • {source.repo_path}"
        )
    else:
        summary = (
            "managed (local source)"
            f" • {source.repo_url}{ref_label}"
            f" • {source.repo_path}"
        )
    return SkillProvenance(status="managed", summary=summary, source=source)


def format_skill_provenance(skill_dir: Path) -> str:
    return get_skill_provenance(skill_dir).summary


def format_revision_short(revision: str | None) -> str:
    return marketplace_formatting.format_revision_short(revision)


def format_installed_at_display(installed_at: str | None) -> str:
    return marketplace_formatting.format_installed_at_display(installed_at)


def format_skill_provenance_details(skill_dir: Path) -> tuple[str, str | None]:
    provenance = get_skill_provenance(skill_dir)
    if provenance.status == "unmanaged":
        return "unmanaged.", None
    if provenance.status != "managed" or provenance.source is None:
        return provenance.summary, None

    source = provenance.source
    ref_label = f"@{source.repo_ref}" if source.repo_ref else ""
    provenance_value = f"{source.repo_url}{ref_label} ({source.repo_path})"
    installed_value = (
        f"{format_installed_at_display(source.installed_at)} "
        f"revision: {format_revision_short(source.installed_revision)}"
    )
    return provenance_value, installed_value


def parse_installed_skill_source_payload(payload: dict[str, Any]) -> InstalledSkillSource:
    parsed = marketplace_source_utils.parse_installed_source_fields(
        payload,
        expected_schema_version=SKILL_SOURCE_SCHEMA_VERSION,
        normalize_repo_path=normalize_repo_path,
    )

    return InstalledSkillSource(
        schema_version=SKILL_SOURCE_SCHEMA_VERSION,
        installed_via="marketplace",
        source_origin=parsed.source_origin,
        repo_url=parsed.repo_url,
        repo_ref=parsed.repo_ref,
        repo_path=parsed.repo_path,
        source_url=parsed.source_url,
        installed_commit=parsed.installed_commit,
        installed_path_oid=parsed.installed_path_oid,
        installed_revision=parsed.installed_revision,
        installed_at=parsed.installed_at,
        content_fingerprint=parsed.content_fingerprint,
    )


def build_installed_skill_source(
    *,
    skill: MarketplaceSkill,
    source_origin: SkillSourceOrigin,
    installed_commit: str | None,
    installed_path_oid: str | None,
    fingerprint: str,
) -> InstalledSkillSource:
    installed_revision = installed_commit or LOCAL_REVISION
    return InstalledSkillSource(
        schema_version=SKILL_SOURCE_SCHEMA_VERSION,
        installed_via="marketplace",
        source_origin=source_origin,
        repo_url=skill.repo_url,
        repo_ref=skill.repo_ref,
        repo_path=skill.repo_path,
        source_url=skill.source_url,
        installed_commit=installed_commit,
        installed_path_oid=installed_path_oid,
        installed_revision=installed_revision,
        installed_at=_iso_utc_now(),
        content_fingerprint=fingerprint,
    )


def _iso_utc_now() -> str:
    return datetime.now(UTC).replace(microsecond=0).isoformat().replace("+00:00", "Z")
